# Unit test set_out_sample_residuals ForecasterRecursiveMultiSeries
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression
from sklearn.exceptions import NotFittedError
from sklearn.preprocessing import StandardScaler
from skforecast.exceptions import UnknownLevelWarning
from ....recursive import ForecasterRecursiveMultiSeries

# Fixtures
series = {
    'l1': pd.Series(np.arange(10)), 
    'l2': pd.Series(np.arange(10))
}


def test_set_out_sample_residuals_NotFittedError_when_forecaster_not_fitted():
    """
    Test NotFittedError is raised when forecaster is not fitted.
    """
    forecaster = ForecasterRecursiveMultiSeries(LinearRegression(), lags=3)
    y_true = {'1': np.array([1, 2, 3, 4, 5]), '2': np.array([1, 2, 3, 4, 5])}
    y_pred = {'1': np.array([1, 2, 3, 4, 5]), '2': np.array([1, 2, 3, 4, 5])}

    err_msg = re.escape(
        "This forecaster is not fitted yet. Call `fit` with appropriate "
        "arguments before using `set_out_sample_residuals()`."
    )
    with pytest.raises(NotFittedError, match = err_msg):
        forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)


def test_set_out_sample_residuals_TypeError_when_y_true_is_not_dict():
    """
    Test TypeError is raised when y_true is not a dict.
    """
    forecaster = ForecasterRecursiveMultiSeries(LinearRegression(), lags=3)
    forecaster.is_fitted = True
    y_true = 'not_dict'
    y_pred = {'1': np.array([1, 2, 3, 4, 5]), '2': np.array([1, 2, 3, 4, 5])}

    err_msg = re.escape(
        f"`y_true` must be a dictionary of numpy ndarrays or pandas Series. "
        f"Got {type(y_true)}."
    )
    with pytest.raises(TypeError, match = err_msg):
        forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)


def test_set_out_sample_residuals_TypeError_when_y_pred_is_not_dict():
    """
    Test TypeError is raised when y_pred is not a dict.
    """
    forecaster = ForecasterRecursiveMultiSeries(LinearRegression(), lags=3)
    forecaster.is_fitted = True
    y_true = {'1': np.array([1, 2, 3, 4, 5]), '2': np.array([1, 2, 3, 4, 5])}
    y_pred = 'not_dict'

    err_msg = re.escape(
        f"`y_pred` must be a dictionary of numpy ndarrays or pandas Series. "
        f"Got {type(y_pred)}."
    )
    with pytest.raises(TypeError, match = err_msg):
        forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)


def test_set_out_sample_residuals_ValueError_when_y_pred_and_y_true_keys_do_not_match():
    """
    Test TypeError is raised when y_pred and y_true keys do not match.
    """
    forecaster = ForecasterRecursiveMultiSeries(LinearRegression(), lags=3)
    forecaster.is_fitted = True
    y_true = {'1': np.array([1, 2, 3, 4, 5]), '2': np.array([1, 2, 3, 4, 5])}
    y_pred = {'3': np.array([1, 2, 3, 4, 5]), '4': np.array([1, 2, 3, 4, 5])}

    err_msg = re.escape(
        f"`y_true` and `y_pred` must have the same keys. "
        f"Got {set(y_true.keys())} and {set(y_pred.keys())}."
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)


def test_set_out_sample_residuals_TypeError_when_y_true_contains_no_numpy_ndarrays_or_pandas_Series():
    """
    Test TypeError is raised when y_true contains no numpy ndarrays or pandas Series.
    """
    forecaster = ForecasterRecursiveMultiSeries(LinearRegression(), lags=3)
    forecaster.is_fitted = True
    y_true = {'1': 'not_ndarray'}
    y_pred = {'1': np.array([1, 2, 3, 4, 5])}

    err_msg = re.escape(
        f"Values of `y_true` must be numpy ndarrays or pandas Series. "
        f"Got {type(y_true['1'])} for series '1'."
    )
    with pytest.raises(TypeError, match = err_msg):
        forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)


def test_set_out_sample_residuals_TypeError_when_y_pred_contains_no_numpy_ndarrays_or_pandas_Series():
    """
    Test TypeError is raised when y_pred contains no numpy ndarrays or pandas Series.
    """
    forecaster = ForecasterRecursiveMultiSeries(LinearRegression(), lags=3)
    forecaster.is_fitted = True
    y_true = {'1': np.array([1, 2, 3, 4, 5])}
    y_pred = {'1': 'not_ndarray'}

    err_msg = re.escape(
        f"Values of `y_pred` must be numpy ndarrays or pandas Series. "
        f"Got {type(y_pred['1'])} for series '1'."
    )
    with pytest.raises(TypeError, match = err_msg):
        forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)


def test_set_out_sample_residuals_ValueError_when_y_true_and_y_pred_have_elements_with_different_lengths():
    """
    Test ValueError is raised when y_true and y_pred have elements with different lengths.
    """
    forecaster = ForecasterRecursiveMultiSeries(LinearRegression(), lags=3)
    forecaster.is_fitted = True
    y_true = {'1': np.array([1, 2, 3, 4, 5]), '2': np.array([1, 2, 3, 4, 5])}
    y_pred = {'1': np.array([1, 2, 3, 4, 5]), '2': np.array([1, 2])}

    err_msg = re.escape(
        f"`y_true` and `y_pred` must have the same length. "
        f"Got {len(y_true['2'])} and {len(y_pred['2'])} for series '2'."
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)


def test_set_out_sample_residuals_ValueError_when_y_true_and_y_pred_have_series_with_different_indexes():
    """
    Test ValueError is raised when y_true and y_pred have series with different indexes.
    """
    forecaster = ForecasterRecursiveMultiSeries(LinearRegression(), lags=3)
    forecaster.is_fitted = True
    y_true = {'1': pd.Series([1, 2, 3, 4, 5], index=[1, 2, 3, 4, 5])}
    y_pred = {'1': pd.Series([1, 2, 3, 4, 5])}

    err_msg = re.escape(
        "When containing pandas Series, elements in `y_true` and "
        "`y_pred` must have the same index. Error with series '1'."
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)


def test_set_out_sample_residuals_ValueError_when_inputs_does_not_match_series_seen_in_fit():
    """
    Test ValueError is raised when inputs does not contain keys that match any 
    series seen in fit.
    """
    forecaster = ForecasterRecursiveMultiSeries(LinearRegression(), lags=3)
    forecaster.fit(series=series)
    y_true = {'5': np.array([1, 2, 3])}
    y_pred = {'5': np.array([1, 2, 3])}

    err_msg = re.escape(
        "Provided keys in `y_pred` and `y_true` do not match any series "
        "seen during `fit`. Residuals cannot be updated."
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)


def test_set_out_sample_residuals_UnknownLevelWarning_when_residuals_levels_but_encoding_None():
    """
    Test UnknownLevelWarning is raised when residuals contains levels but encoding is None.
    """
    forecaster = ForecasterRecursiveMultiSeries(LinearRegression(), lags=3, encoding=None)
    forecaster.fit(series=series)
    y_true = {'l1': np.array([1, 2, 3, 4, 5])}
    y_pred = {'l1': np.array([1, 2, 3, 4, 5])}

    err_msg = re.escape(
        "As `encoding` is set to `None`, no distinction between levels "
        "is made. All residuals are stored in the '_unknown_level' key."
    )
    with pytest.warns(UnknownLevelWarning, match = err_msg):
        forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)
    results = forecaster.out_sample_residuals_

    expected = {
        '_unknown_level': np.array([0, 0, 0, 0, 0])
    }

    assert expected.keys() == results.keys()
    for k in results.keys():
        np.testing.assert_array_almost_equal(expected[k], results[k])


@pytest.mark.parametrize("encoding", 
                         ['ordinal', 'onehot', 'ordinal_category', None], 
                         ids=lambda encoding: f'encoding: {encoding}')
def test_set_out_sample_residuals_when_residuals_length_is_less_than_10000_and_no_append(encoding):
    """
    Test residuals stored when new residuals length is less than 10000 and append
    is False.
    """
    rng = np.random.default_rng(12345)
    series = {
        "l1": pd.Series(rng.normal(10, 3, 20)), 
        "l2": pd.Series(rng.normal(10, 3, 20))
    }
    forecaster = ForecasterRecursiveMultiSeries(
        LinearRegression(), lags=3, encoding=encoding, binner_kwargs={"n_bins": 3}
    )
    forecaster.fit(series=series, store_in_sample_residuals=True)
    y_true = {"l1": rng.normal(10, 3, 20), "l2": rng.normal(10, 3, 20)}
    y_pred = {"l1": rng.normal(10, 3, 20), "l2": rng.normal(10, 3, 20)}

    forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)
    expected_out_sample = {
        'l1': np.array(
            [1.60962553,  1.71263736,  9.19708889, -2.73386424,  1.51464604,
            -3.32699998,  3.5987987 ,  0.13608979, -2.06018676, -2.72988393,
            1.63221968, -0.57130461, -1.08106429, -1.45886519,  0.91789988,
            -2.19933831,  1.97853178, -2.56217282,  3.69683321, -4.86497532]),
        'l2': np.array(
                [0.51984518,  7.40971902, -0.77881843, -2.89834312, -4.09573881,
                2.98823132,  2.43787115, -5.81851232,  1.34536138, -2.88085653,
                -8.26701406, -8.9861346 ,  0.31671219, -0.83725661, -4.37490671,
                -2.88874692,  0.50604084, -3.41273397, -2.71056093,  3.120254 ]),
        '_unknown_level':  np.array(
                [1.60962553,  1.71263736,  9.19708889, -2.73386424,  1.51464604,
                -3.32699998,  3.5987987 ,  0.13608979, -2.06018676, -2.72988393,
                1.63221968, -0.57130461, -1.08106429, -1.45886519,  0.91789988,
                -2.19933831,  1.97853178, -2.56217282,  3.69683321, -4.86497532,
                0.51984518,  7.40971902, -0.77881843, -2.89834312, -4.09573881,
                2.98823132,  2.43787115, -5.81851232,  1.34536138, -2.88085653,
                -8.26701406, -8.9861346 ,  0.31671219, -0.83725661, -4.37490671,
                -2.88874692,  0.50604084, -3.41273397, -2.71056093,  3.120254])
            }

    expected_out_sample_by_bin = {
        "l1": {
            0: np.array([
                1.60962553, 1.71263736, 9.19708889, 1.51464604, 3.5987987,
                1.63221968, -2.19933831, 1.97853178, -2.56217282, 3.69683321,
            ]),
            1: np.array([
                -3.32699998,
            ]),
            2: np.array([
                -2.73386424, 0.13608979, -2.06018676, -2.72988393, -0.57130461,
                -1.08106429, -1.45886519, 0.91789988, -4.86497532,
            ]),
        },
        "l2": {
            0: np.array([
                0.51984518, 7.40971902, -4.09573881, 2.98823132, -2.88085653,
                3.120254,
            ]),
            1: np.array([
                -0.77881843, 2.43787115, 0.31671219, -2.71056093,
            ]),
            2: np.array([
                -2.89834312, -5.81851232, 1.34536138, -8.26701406, -8.9861346,
                -0.83725661, -4.37490671, -2.88874692, 0.50604084, -3.41273397,
            ]),
        },
        "_unknown_level": {
            0: np.array([
                1.60962553, 1.71263736, 9.19708889, 1.51464604, 3.5987987,
                1.63221968, -2.19933831, 1.97853178, -2.56217282, 3.69683321,
                0.51984518, 7.40971902, -4.09573881, 2.98823132, -2.88085653,
                3.120254,
            ]),
            1: np.array([
                -3.32699998, -0.77881843, 2.43787115, 0.31671219, -2.71056093,
            ]),
            2: np.array([
                -2.73386424, 0.13608979, -2.06018676, -2.72988393, -0.57130461,
                -1.08106429, -1.45886519, 0.91789988, -4.86497532, -2.89834312,
                -5.81851232, 1.34536138, -8.26701406, -8.9861346, -0.83725661,
                -4.37490671, -2.88874692, 0.50604084, -3.41273397,
            ]),
        },
    }

    if encoding is None:
        expected_out_sample = {'_unknown_level': expected_out_sample['_unknown_level']}
        expected_out_sample_by_bin = {
            '_unknown_level': {
                0: np.array([ 1.60962553,  1.71263736,  9.19708889,  1.51464604,  3.5987987 ,
                    1.97853178, -2.56217282,  3.69683321,  0.51984518,  7.40971902,
                    -0.77881843, -4.09573881,  2.98823132,  2.43787115, -2.88085653,
                    3.120254  ]),
                1: np.array([-3.32699998,  1.63221968, -2.19933831,  0.31671219, -2.71056093]),
                2: np.array([-2.73386424,  0.13608979, -2.06018676, -2.72988393, -0.57130461,
                    -1.08106429, -1.45886519,  0.91789988, -4.86497532, -2.89834312,
                    -5.81851232,  1.34536138, -8.26701406, -8.9861346 , -0.83725661,
                    -4.37490671, -2.88874692,  0.50604084, -3.41273397])
            }
        }
        
    assert expected_out_sample.keys() == forecaster.out_sample_residuals_.keys()
    for k in expected_out_sample.keys():
        np.testing.assert_array_almost_equal(
            np.sort(expected_out_sample[k]), 
            np.sort(forecaster.out_sample_residuals_[k])
        )

    assert expected_out_sample_by_bin.keys() == forecaster.out_sample_residuals_by_bin_.keys()
    for k in expected_out_sample_by_bin.keys():
        for bin in expected_out_sample_by_bin[k].keys():
            np.testing.assert_array_almost_equal(
                np.sort(expected_out_sample_by_bin[k][bin]),
                np.sort(forecaster.out_sample_residuals_by_bin_[k][bin])
            )


@pytest.mark.parametrize("encoding", 
                         ['ordinal', 'onehot', 'ordinal_category', None], 
                         ids=lambda encoding: f'encoding: {encoding}')
def test_set_out_sample_residuals_for_unknown_level(encoding):
    """
    Test residuals stored for unknown level.
    """
    forecaster = ForecasterRecursiveMultiSeries(LinearRegression(), lags=3, encoding=encoding)
    forecaster.fit(series=series)
    y_true = {'_unknown_level': np.array([1, 2, 3, 5, 6])}
    y_pred = {'_unknown_level': np.array([0, 1, 2, 3, 4])}

    forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)
    results = forecaster.out_sample_residuals_

    if encoding is None:
        expected = {
            '_unknown_level': np.array([1, 1, 1, 2, 2])
        }
    else:
        expected = {
            'l1': None,
            'l2': None,
            '_unknown_level': np.array([1, 1, 1, 2, 2])
        }

    assert expected.keys() == results.keys()
    for k in results.keys():
        if results[k] is None:
            assert results[k] == expected[k]
        else:
            np.testing.assert_array_almost_equal(expected[k], results[k])


def test_set_out_sample_residuals_when_residuals_length_is_less_than_10000_encoding_None():
    """
    Test residuals stored when new residuals length is less than 10000 and 
    encoding is None.
    """
    forecaster = ForecasterRecursiveMultiSeries(
                    LinearRegression(),
                    lags=3,
                    encoding=None,
                 )
    forecaster.fit(series=series)
    y_true = {'l1': np.array([1, 2, 3, 4, 5]), 'l2': np.array([2, 3, 4, 5, 6])}
    y_pred = {'l1': np.array([0, 1, 2, 3, 4]), 'l2': np.array([0, 1, 2, 3, 4])}

    forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)
    results = forecaster.out_sample_residuals_
    expected = {
        '_unknown_level': np.array([2, 2, 2, 2, 2, 1, 1, 1, 1, 1])
    }

    assert expected.keys() == results.keys()
    assert all(all(np.sort(expected[k]) == np.sort(results[k])) for k in expected.keys())


@pytest.mark.parametrize("encoding", ['ordinal', 'onehot', 'ordinal_category'], 
                         ids=lambda encoding: f'encoding: {encoding}')
def test_set_out_sample_residuals_when_residuals_length_is_less_than_10000_and_append(encoding):
    """
    Test residuals stored when new residuals length is less than 10000 and append is True.
    """
    forecaster = ForecasterRecursiveMultiSeries(
                    LinearRegression(),
                    lags=3,
                    encoding=encoding,
                 )
    forecaster.fit(series=series)
    y_true = {'l1': np.array([1, 2, 3, 4, 5]), 'l2': np.array([2, 3, 4, 5, 6])}
    y_pred = {'l1': np.array([0, 1, 2, 3, 4]), 'l2': np.array([0, 1, 2, 3, 4])}
    
    forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)
    forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred, append=True)
    results = forecaster.out_sample_residuals_

    expected = {
        'l1': np.array([1, 1, 1, 1, 1] * 2), 
        'l2': np.array([2, 2, 2, 2, 2] * 2),
        '_unknown_level': np.concatenate([
            np.array([1, 1, 1, 1, 1] * 2),
            np.array([2, 2, 2, 2, 2] * 2)

        ])
    }

    assert expected.keys() == results.keys()
    for k in results.keys():
        np.testing.assert_array_almost_equal(np.sort(expected[k]), np.sort(results[k]))


@pytest.mark.parametrize("encoding", 
                         ['ordinal', 'onehot', 'ordinal_category'], 
                         ids=lambda encoding: f'encoding: {encoding}')
def test_set_out_sample_residuals_when_residuals_length_is_greater_than_10000(encoding):
    """
    Test len residuals stored when its length is greater than 10000.
    """
    forecaster = ForecasterRecursiveMultiSeries(
        LinearRegression(), lags=3, encoding=encoding, binner_kwargs={'n_bins': 2}
    )
    forecaster.fit(series=series)
    y_true = {'l1': np.ones(20_000), 'l2': np.ones(20_000)}
    y_pred = {
        'l1': np.concatenate([np.ones(10_000) + 1, np.ones(10_000) + 2]),
        'l2': np.concatenate([np.ones(10_000) + 1, np.ones(10_000) + 2])
    }
    forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)
    results = forecaster.out_sample_residuals_, forecaster.out_sample_residuals_by_bin_

    assert list(results[0].keys()) == ['l1', 'l2', '_unknown_level']
    assert all(len(value) == 10_000 for value in results[0].values())
    for k in results[1].keys():
        for bin in results[1][k].keys():
            if k == '_unknown_level':
                assert len(results[1][k][bin]) == 10_000
            else:
                assert len(results[1][k][bin]) == 5_000


def test_set_out_sample_residuals_when_residuals_length_is_greater_than_10000_encoding_None():
    """
    Test len residuals stored when its length is greater than 10_000
    and encoding is None.
    """
    forecaster = ForecasterRecursiveMultiSeries(
        LinearRegression(), lags=3, encoding=None, binner_kwargs={'n_bins': 2}
    )
    forecaster.fit(series=series)
    y_true = {'l1': np.ones(20_000), 'l2': np.ones(20_000)}
    y_pred = {
        'l1': np.concatenate([np.ones(10_000) + 1, np.ones(10_000) + 2]),
        'l2': np.concatenate([np.ones(10_000) + 1, np.ones(10_000) + 2])
    }
    forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)
    results = forecaster.out_sample_residuals_, forecaster.out_sample_residuals_by_bin_

    assert list(results[0].keys()) == ['_unknown_level']
    assert all(len(value) == 10_000 for value in results[0].values())
    for k in results[1].keys():
        for bin in results[1][k].keys():
            assert len(results[1][k][bin]) == 10_000


def test_set_out_sample_residuals_when_residuals_length_is_greater_than_10000_and_append():
    """
    Test residuals stored when new residuals length is greater than 10000 and
    append is True.
    """
    forecaster = ForecasterRecursiveMultiSeries(
        LinearRegression(), lags=3, encoding='ordinal',  binner_kwargs={'n_bins': 2}
    )
    forecaster.fit(series=series)
    y_true = {'l1': np.ones(5_000), 'l2': np.ones(5_000)}
    y_pred = {
        'l1': np.concatenate([np.ones(2_500) + 1, np.ones(2_500) + 2]), 
        'l2': np.concatenate([np.ones(2_500) + 1, np.ones(2_500) + 2])
    }
    forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)

    y_true = {'l1': np.ones(20_000), 'l2': np.ones(20_000)}
    y_pred = {
        'l1': np.concatenate([np.ones(10_000) + 1, np.ones(10_000) + 2]), 
        'l2': np.concatenate([np.ones(10_000) + 1, np.ones(10_000) + 2])
    }
    forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred, append=True)
    results = forecaster.out_sample_residuals_, forecaster.out_sample_residuals_by_bin_

    assert list(results[0].keys()) == ['l1', 'l2', '_unknown_level']
    assert all([len(v) == 10_000 for v in results[0].values()])
    for k in results[1].keys():
        for bin in results[1][k].keys():
            if k == '_unknown_level':
                assert len(results[1][k][bin]) == 10_000
            else:
                assert len(results[1][k][bin]) == 5_000


def test_set_out_sample_residuals_when_residuals_keys_partially_match():
    """
    Test residuals are stored only for matching keys.
    """
    forecaster = ForecasterRecursiveMultiSeries(
                    LinearRegression(),
                    lags=3,
                 )
    forecaster.fit(series=series)
    y_pred = {'l1': np.repeat(1, 5), 'l4': np.arange(10)}
    y_true = {'l1': np.arange(5), 'l4': np.arange(10)}
    forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)
    results = forecaster.out_sample_residuals_
    expected = {
        'l1': np.array([-1,  0,  1,  2,  3]),
        'l2': None,
        '_unknown_level': np.array([-1, 0, 1, 2, 3])
    }
    for key in expected.keys():
        if expected[key] is not None:
            np.testing.assert_array_almost_equal(expected[key], results[key])
        else:
            assert results[key] is None


@pytest.mark.parametrize("differentiation", 
                         [1, {'l1': 1, 'l2': 1, '_unknown_level': 1}], 
                         ids=lambda diff: f'differentiation: {diff}')
def test_forecaster_set_out_sample_residuals_when_transformer_series_and_differentiation(differentiation):
    """
    Test set_out_sample_residuals when forecaster has transformer_series and differentiation.
    Stored should equivalent to residuals calculated manually if transformer_series and
    differentiation are applied to `y_true` and `y_pred` before calculating residuals.
    """
    series_train = {
        'l1': pd.Series(
            np.array([-1.42382504,  1.26372846, -0.87066174, -0.25917323, -0.07534331,
                      -0.74088465, -1.3677927 ,  0.6488928 ,  0.36105811, -1.95286306,
                       2.34740965,  0.96849691, -0.75938718,  0.90219827, -0.46695317,
                      -0.06068952,  0.78884434, -1.25666813,  0.57585751,  1.39897899]),
            index = pd.date_range(start='1-1-2018', periods=20, freq='D')
        ),
        'l2': pd.Series(
            np.array([1.32229806, -0.29969852,  0.90291934, -1.62158273, -0.15818926,
                      0.44948393, -1.34360107, -0.08168759,  1.72473993,  2.61815943,
                      0.77736134,  0.8286332 , -0.95898831, -1.20938829, -1.41229201,
                      0.54154683,  0.7519394 , -0.65876032, -1.22867499,  0.25755777]),
            index = pd.date_range(start='1-1-2018', periods=20, freq='D')
        )
    }
    y_true  = {
        'l1': np.array([ 0.31290292, -0.13081169,  1.26998312, -0.09296246, -0.06615089]),
        'l2': np.array([-1.10821447,  0.13595685,  1.34707776,  0.06114402,  0.0709146 ])
    }
    y_pred = {
        'l1': np.array([0.43365454, 0.27748366, 0.53025239, 0.53672097, 0.61835001]),
        'l2': np.array([-0.79501746,  0.30003095, -1.60270159,  0.26679883, -1.26162378])
    }
    
    forecaster = ForecasterRecursiveMultiSeries(
                     regressor          = LinearRegression(),
                     lags               = 5,
                     transformer_series = StandardScaler(),
                     differentiation    = differentiation
                 )
    forecaster.fit(series=series_train)
    forecaster.set_out_sample_residuals(
        y_true = y_true,
        y_pred = y_pred
    )

    y_true['l1'] = forecaster.transformer_series_['l1'].transform(y_true['l1'].reshape(-1, 1)).flatten()
    y_true['l2'] = forecaster.transformer_series_['l2'].transform(y_true['l2'].reshape(-1, 1)).flatten()
    y_pred['l1'] = forecaster.transformer_series_['l1'].transform(y_pred['l1'].reshape(-1, 1)).flatten()
    y_pred['l2'] = forecaster.transformer_series_['l2'].transform(y_pred['l2'].reshape(-1, 1)).flatten()
    y_true['l1'] = forecaster.differentiator_['l1'].transform(y_true['l1'])[forecaster.differentiation_max:]
    y_true['l2'] = forecaster.differentiator_['l2'].transform(y_true['l2'])[forecaster.differentiation_max:]
    y_pred['l1'] = forecaster.differentiator_['l1'].transform(y_pred['l1'])[forecaster.differentiation_max:]
    y_pred['l2'] = forecaster.differentiator_['l2'].transform(y_pred['l2'])[forecaster.differentiation_max:]
    residuals = {}
    residuals['l1'] = y_true['l1'] - y_pred['l1']
    residuals['l2'] = y_true['l2'] - y_pred['l2']
    residuals['_unknown_level'] = np.sort(np.concatenate([(y_true['l2'] - y_pred['l2']), (y_true['l1'] - y_pred['l1'])]))
    forecaster.out_sample_residuals_['_unknown_level'] = np.sort(forecaster.out_sample_residuals_['_unknown_level'])

    for key in residuals.keys():
        np.testing.assert_array_almost_equal(residuals[key], forecaster.out_sample_residuals_[key])


def test_forecaster_set_out_sample_residuals_when_transformer_series_and_differentiation_as_dict_unknonw_level():
    """
    Test set_out_sample_residuals when forecaster has transformer_series, differentiation
    and unknown level.
    Stored should equivalent to residuals calculated manually if transformer_series and
    differentiation are applied to `y_true` and `y_pred` before calculating residuals.
    """
    rng = np.random.default_rng(12345)
    series_train = {
        'l1': pd.Series(
            rng.normal(loc=0, scale=1, size=100),
            index = pd.date_range(start='1-1-2018', periods=100, freq='D')
        ),
        'l2': pd.Series(
            rng.normal(loc=0, scale=1, size=100),
            index = pd.date_range(start='1-1-2018', periods=100, freq='D')
        ),
        'l3': pd.Series(
            rng.normal(loc=0, scale=1, size=100),
            index = pd.date_range(start='1-1-2018', periods=100, freq='D')
        )
    }
    y_true  = {
        'l1': rng.normal(loc=0, scale=1, size=5),
        'l2': rng.normal(loc=0, scale=1, size=5),
        'l3': rng.normal(loc=0, scale=1, size=5),
        '_unknown_level': rng.normal(loc=0, scale=1, size=5)
    }
    y_pred = {
        'l1': rng.normal(loc=0, scale=1, size=5),
        'l2': rng.normal(loc=0, scale=1, size=5),
        'l3': rng.normal(loc=0, scale=1, size=5),
        '_unknown_level': rng.normal(loc=0, scale=1, size=5)
    }
    forecaster = ForecasterRecursiveMultiSeries(
                     regressor          = LinearRegression(),
                     lags               = 5,
                     differentiation    = {'l1': 1, 'l2': 2, 'l3': None, '_unknown_level': 1},
                     transformer_series = StandardScaler(),
                 )
    forecaster.fit(series=series_train)
    forecaster.set_out_sample_residuals(
        y_true = y_true,
        y_pred = y_pred
    )

    y_true['l1'] = forecaster.transformer_series_['l1'].transform(y_true['l1'].reshape(-1, 1)).flatten()
    y_true['l2'] = forecaster.transformer_series_['l2'].transform(y_true['l2'].reshape(-1, 1)).flatten()
    y_true['l3'] = forecaster.transformer_series_['l3'].transform(y_true['l3'].reshape(-1, 1)).flatten()
    y_true['_unknown_level'] = forecaster.transformer_series_['_unknown_level'].transform(y_true['_unknown_level'].reshape(-1, 1)).flatten()
    y_pred['l1'] = forecaster.transformer_series_['l1'].transform(y_pred['l1'].reshape(-1, 1)).flatten()
    y_pred['l2'] = forecaster.transformer_series_['l2'].transform(y_pred['l2'].reshape(-1, 1)).flatten()
    y_pred['l3'] = forecaster.transformer_series_['l3'].transform(y_pred['l3'].reshape(-1, 1)).flatten()
    y_pred['_unknown_level'] = forecaster.transformer_series_['_unknown_level'].transform(y_pred['_unknown_level'].reshape(-1, 1)).flatten()
    y_true['l1'] = forecaster.differentiator_['l1'].transform(y_true['l1'])[forecaster.differentiator_['l1'].order:]
    y_true['l2'] = forecaster.differentiator_['l2'].transform(y_true['l2'])[forecaster.differentiator_['l2'].order:]
    # l3 is not differentiated
    y_true['_unknown_level'] = forecaster.differentiator_['_unknown_level'].transform(y_true['_unknown_level'])[forecaster.differentiator_['_unknown_level'].order:]
    y_pred['l1'] = forecaster.differentiator_['l1'].transform(y_pred['l1'])[forecaster.differentiator_['l1'].order:]
    y_pred['l2'] = forecaster.differentiator_['l2'].transform(y_pred['l2'])[forecaster.differentiator_['l2'].order:]
    # l3 is not differentiated
    y_pred['_unknown_level'] = forecaster.differentiator_['_unknown_level'].transform(y_pred['_unknown_level'])[forecaster.differentiator_['_unknown_level'].order:]
    residuals = {}
    residuals['l1'] = y_true['l1'] - y_pred['l1']
    residuals['l2'] = y_true['l2'] - y_pred['l2']
    residuals['l3'] = y_true['l3'] - y_pred['l3']
    residuals['_unknown_level'] = y_true['_unknown_level'] - y_pred['_unknown_level']

    for key in residuals.keys():
        np.testing.assert_array_almost_equal(residuals[key], forecaster.out_sample_residuals_[key])



# Unit test skforecast version
# ==============================================================================
from skforecast import __version__
import tomli

version = "0.17.0"


def test_version_in_init_and_pyproject_toml():
    """
    Test that version in __init__.py and pyproject.toml are the same.

    Parameters
    ----------
    version : str, optional
    """

    with open("./pyproject.toml", mode='rb') as fp:
        pyproject = tomli.load(fp)

    version_in_tolm = pyproject['project']['version']
    version_in_init = __version__

    assert version_in_init == version
    assert version_in_tolm == version    


# Test are located inside each module's folder


# Unit test set_window_features ForecasterDirect
# ==============================================================================
import re
import pytest
from sklearn.linear_model import LinearRegression 
from skforecast.preprocessing import RollingFeatures
from skforecast.direct import ForecasterDirect


def test_set_window_features_ValueError_when_window_features_set_to_None_and_lags_is_None():
    """
    Test ValueError is raised when window_features is set to None and lags is None.
    """
    rolling = RollingFeatures(stats='mean', window_sizes=6)
    forecaster = ForecasterDirect(
        LinearRegression(), steps=2, lags=None, window_features=rolling
    )

    err_msg = re.escape(
        "At least one of the arguments `lags` or `window_features` "
        "must be different from None. This is required to create the "
        "predictors used in training the forecaster."
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.set_window_features(window_features=None)


@pytest.mark.parametrize("wf", 
                         [RollingFeatures(stats='mean', window_sizes=6),
                          [RollingFeatures(stats='mean', window_sizes=6)]],
                         ids = lambda wf: f'window_features: {type(wf)}')
def test_set_window_features_with_different_inputs(wf):
    """
    Test how attributes change with window_features argument.
    """
    forecaster = ForecasterDirect(LinearRegression(), lags=5, steps=2)
    forecaster.set_window_features(window_features=wf)

    assert forecaster.lags_names == ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5']
    assert forecaster.max_lag == 5
    assert forecaster.max_size_window_features == 6
    assert forecaster.window_features_names == ['roll_mean_6']
    assert forecaster.window_features_class_names == ['RollingFeatures']
    assert forecaster.window_size == 6


def test_set_window_features_when_differentiation_is_not_None():
    """
    Test how `window_size` is also updated when the forecaster includes 
    differentiation.
    """
    forecaster = ForecasterDirect(
                     regressor       = LinearRegression(),
                     steps           = 2,
                     lags            = 3,
                     window_features = RollingFeatures(stats='median', window_sizes=2),
                     differentiation = 1
                 )
    
    assert forecaster.max_size_window_features == 2
    assert forecaster.window_features_names == ['roll_median_2']
    assert forecaster.window_features_class_names == ['RollingFeatures']
    assert forecaster.window_size == 3 + 1
    assert forecaster.differentiator.window_size == 3 + 1
    
    rolling = RollingFeatures(stats='mean', window_sizes=6)
    forecaster.set_window_features(window_features=rolling)

    assert forecaster.lags_names == ['lag_1', 'lag_2', 'lag_3']
    assert forecaster.max_lag == 3
    assert forecaster.max_size_window_features == 6
    assert forecaster.window_features_names == ['roll_mean_6']
    assert forecaster.window_features_class_names == ['RollingFeatures']
    assert forecaster.window_size == 6 + 1
    assert forecaster.differentiator.window_size == 6 + 1


def test_set_window_features_when_lags():
    """
    Test how `window_size` is also updated when the forecaster includes
    lags.
    """
    rolling = RollingFeatures(stats='mean', window_sizes=10)
    forecaster = ForecasterDirect(
                     regressor       = LinearRegression(),
                     lags            = 9,
                     steps           = 2,
                     window_features = rolling
                 )
    
    rolling = RollingFeatures(stats='median', window_sizes=5)
    forecaster.set_window_features(window_features=rolling)

    assert forecaster.lags_names == ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5',
                                     'lag_6', 'lag_7', 'lag_8', 'lag_9']
    assert forecaster.max_lag == 9
    assert forecaster.max_size_window_features == 5
    assert forecaster.window_features_names == ['roll_median_5']
    assert forecaster.window_features_class_names == ['RollingFeatures']
    assert forecaster.window_size == 9


def test_set_window_features_to_None():
    """
    Test how attributes change when window_features is set to None.
    """
    rolling = RollingFeatures(stats='mean', window_sizes=6)
    forecaster = ForecasterDirect(
                     regressor       = LinearRegression(),
                     lags            = 5,
                     steps           = 3,
                     window_features = rolling
                 )
    
    forecaster.set_window_features(window_features=None)
    
    assert forecaster.lags_names == ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5']
    assert forecaster.max_lag == 5
    assert forecaster.max_size_window_features is None
    assert forecaster.window_features_names is None
    assert forecaster.window_features_class_names is None
    assert forecaster.window_size == 5



# Unit test get_exog_dtypes
# ==============================================================================
import pandas as pd
from skforecast.utils import get_exog_dtypes


def test_get_exog_dtypes_when_pandas_Series():
    """
    Check get_exog_dtypes when exog is pandas Series.
    """
    exog = pd.Series([1, 2, 3], name='y')
    exog_dtypes = get_exog_dtypes(exog)

    expected = {'y': exog.dtypes}

    assert exog_dtypes['y'] == expected['y']


def test_get_exog_dtypes_when_pandas_DataFrame():
    """
    Check get_exog_dtypes when exog is pandas DataFrame.
    """
    exog = pd.DataFrame({
        'A': [1, 2, 3], 
        'B': [1.1, 2.2, 3.3],
        'C': pd.Categorical(range(3)),
        'D': [True]*3,
        'E': ['string']*3
    })
    exog_dtypes = get_exog_dtypes(exog)

    expected = {
        'A': exog['A'].dtypes,
        'B': exog['B'].dtypes,
        'C': exog['C'].dtypes,
        'D': exog['D'].dtypes,
        'E': exog['E'].dtypes
    }

    for k in exog_dtypes.keys():
        assert exog_dtypes[k] == expected[k]


# Unit test bayesian_search_forecaster
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from sklearn.linear_model import Ridge
# from skopt.space import Real
from skforecast.recursive import ForecasterRecursive
from skforecast.preprocessing import RollingFeatures
from skforecast.model_selection._search import bayesian_search_forecaster
from skforecast.model_selection._split import TimeSeriesFold

# Fixtures
from ..fixtures_model_selection import y


def test_ValueError_bayesian_search_forecaster_when_return_best_and_len_y_exog_different():
    """
    Test ValueError is raised in bayesian_search_forecaster when return_best 
    and length of `y` and `exog` do not match.
    """
    forecaster = ForecasterRecursive(
                    regressor = Ridge(random_state=123),
                    lags      = 2
                 )
    exog = y[:30]
    cv = TimeSeriesFold(
            steps                 = 3,
            initial_train_size    = len(y[:-12]),
            window_size           = None,
            differentiation       = None,
            refit                 = True,
            fixed_train_size      = True,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    
    def search_space(trial):  # pragma: no cover
        search_space  = {'alpha': trial.suggest_float('alpha', 1e-2, 1.0)}
        return search_space

    err_msg = re.escape(
        f"`exog` must have same number of samples as `y`. "
        f"length `exog`: ({len(exog)}), length `y`: ({len(y)})"
    )
    with pytest.raises(ValueError, match = err_msg):
        bayesian_search_forecaster(
            forecaster         = forecaster,
            y                  = y,
            exog               = exog,
            cv                 = cv,
            search_space       = search_space,
            metric             = 'mean_absolute_error',
            n_trials           = 10,
            random_state       = 123,
            return_best        = True,
            verbose            = False,
        )


def test_results_output_bayesian_search_forecaster_optuna_ForecasterRecursive_with_mocked():
    """
    Test output of bayesian_search_forecaster in ForecasterRecursive with 
    mocked using optuna (mocked done in Skforecast v0.4.3).
    """    
    forecaster = ForecasterRecursive(
                     regressor = Ridge(random_state=123),
                     lags      = 4
                 )

    cv = TimeSeriesFold(
            steps                 = 3,
            initial_train_size    = len(y[:-12]),
            window_size           = None,
            differentiation       = None,
            refit                 = True,
            fixed_train_size      = True,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    
    def search_space(trial):  # pragma: no cover
        search_space  = {'alpha': trial.suggest_float('alpha', 1e-2, 1.0)}
        return search_space

    results = bayesian_search_forecaster(
                  forecaster         = forecaster,
                  y                  = y,
                  cv                 = cv,
                  search_space       = search_space,
                  metric             = 'mean_absolute_error',
                  n_trials           = 10,
                  random_state       = 123,
                  return_best        = False,
                  verbose            = False,
              )[0]
    
    expected_results = pd.DataFrame(
        np.array([[np.array([1, 2, 3, 4]), {'alpha': 0.9809565564007693},
            0.21539791166603497, 0.9809565564007693],
        [np.array([1, 2, 3, 4]), {'alpha': 0.7222742800877074},
            0.21576131952657338, 0.7222742800877074],
        [np.array([1, 2, 3, 4]), {'alpha': 0.6995044937418831},
            0.21579460210585208, 0.6995044937418831],
        [np.array([1, 2, 3, 4]), {'alpha': 0.6879814411990146},
            0.21581150916013206, 0.6879814411990146],
        [np.array([1, 2, 3, 4]), {'alpha': 0.5558016213920624},
            0.21600778429729228, 0.5558016213920624],
        [np.array([1, 2, 3, 4]), {'alpha': 0.48612258246951734},
            0.21611205459571634, 0.48612258246951734],
        [np.array([1, 2, 3, 4]), {'alpha': 0.42887539552321635},
            0.2161973389956996, 0.42887539552321635],
        [np.array([1, 2, 3, 4]), {'alpha': 0.398196343012209},
            0.2162426532005299, 0.398196343012209],
        [np.array([1, 2, 3, 4]), {'alpha': 0.29327794160087567},
            0.2163933942116072, 0.29327794160087567],
        [np.array([1, 2, 3, 4]), {'alpha': 0.2345829390285611},
            0.21647289061896782, 0.2345829390285611]], dtype=object),
        columns=['lags', 'params', 'mean_absolute_error', 'alpha'],
        index=pd.Index([0, 1, 2, 3, 4, 5, 6, 7, 8, 9], dtype='int64')
    ).astype({
        'mean_absolute_error': float, 
        'alpha': float
    })

    pd.testing.assert_frame_equal(results, expected_results)


def test_results_output_bayesian_search_forecaster_optuna_ForecasterRecursive_window_features_with_mocked():
    """
    Test output of bayesian_search_forecaster in ForecasterRecursive with window features 
    using mocked using optuna (mocked done in Skforecast v0.4.3).
    """
    window_features = RollingFeatures(
        stats        = ['mean', 'std', 'min', 'max', 'sum', 'median', 'ratio_min_max', 'coef_variation'],
        window_sizes = 3,
    )
    forecaster = ForecasterRecursive(
                     regressor       = Ridge(random_state=123),
                     lags            = 4,
                     window_features = window_features,
                 )

    cv = TimeSeriesFold(
            steps                 = 3,
            initial_train_size    = len(y[:-12]),
            window_size           = None,
            differentiation       = None,
            refit                 = True,
            fixed_train_size      = True,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    
    def search_space(trial):  # pragma: no cover
        search_space  = {'alpha': trial.suggest_float('alpha', 1e-2, 1.0)}
        return search_space

    results = bayesian_search_forecaster(
                  forecaster         = forecaster,
                  y                  = y,
                  cv                 = cv,
                  search_space       = search_space,
                  metric             = 'mean_absolute_error',
                  n_trials           = 10,
                  random_state       = 123,
                  return_best        = False,
                  verbose            = False,
              )[0]
    
    expected_results = pd.DataFrame({
            'lags': [
                np.array([1, 2, 3, 4]),
                np.array([1, 2, 3, 4]),
                np.array([1, 2, 3, 4]),
                np.array([1, 2, 3, 4]),
                np.array([1, 2, 3, 4]),
                np.array([1, 2, 3, 4]),
                np.array([1, 2, 3, 4]),
                np.array([1, 2, 3, 4]),
                np.array([1, 2, 3, 4]),
                np.array([1, 2, 3, 4]),
            ],
            'params': [
                {'alpha': 0.9809565564007693},
                {'alpha': 0.7222742800877074},
                {'alpha': 0.6995044937418831},
                {'alpha': 0.6879814411990146},
                {'alpha': 0.5558016213920624},
                {'alpha': 0.48612258246951734},
                {'alpha': 0.42887539552321635},
                {'alpha': 0.398196343012209},
                {'alpha': 0.29327794160087567},
                {'alpha': 0.2345829390285611},
            ],
            'mean_absolute_error': [
                0.23783372219201282,
                0.24038797813509474,
                0.24065731430061374,
                0.24079724452316387,
                0.24261647067658373,
                0.24378706429210686,
                0.24490876891227564,
                0.24558757754181923,
                0.24851969600385504,
                0.2508328027649236,
            ],
            'alpha': [
                0.9809565564007693,
                0.7222742800877074,
                0.6995044937418831,
                0.6879814411990146,
                0.5558016213920624,
                0.48612258246951734,
                0.42887539552321635,
                0.398196343012209,
                0.29327794160087567,
                0.2345829390285611,
            ],
    })

    pd.testing.assert_frame_equal(results, expected_results)


# Unit test initialize_transformer_series
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from sklearn.preprocessing import StandardScaler
from skforecast.utils import initialize_transformer_series
from skforecast.exceptions import IgnoredArgumentWarning


@pytest.mark.parametrize(
    "forecaster_name, encoding, expected", 
    [('ForecasterRecursiveMultiSeries', 'onehot', {'col1': None, 'col2': None, '_unknown_level': None}),
     ('ForecasterRecursiveMultiSeries', None, {'_unknown_level': None}),
     ('ForecasterDirectMultiVariate', None, {'col1': None, 'col2': None}),
     ('ForecasterRNN', None, {'col1': None, 'col2': None})], 
    ids=lambda params: f'params: {params}')
def test_initialize_transformer_series_when_transformer_series_is_None(forecaster_name, encoding, expected):
    """
    Test initialize_transformer_series when `transformer_series` is None.
    """
    series_names_in_ = ['col1', 'col2']
    transformer_series = None
    
    transformer_series_ = initialize_transformer_series(
                              forecaster_name    = forecaster_name,
                              series_names_in_   = series_names_in_,
                              encoding           = encoding,
                              transformer_series = transformer_series
                          )

    assert transformer_series_ == expected


@pytest.mark.parametrize(
    "forecaster_name, encoding, expected", 
    [('ForecasterRecursiveMultiSeries', 'onehot', {'col1': StandardScaler(), 'col2': StandardScaler(), '_unknown_level': StandardScaler()}),
     ('ForecasterRecursiveMultiSeries', None, {'_unknown_level': StandardScaler()}),
     ('ForecasterDirectMultiVariate', None, {'col1': StandardScaler(), 'col2': StandardScaler()}),
     ('ForecasterRNN', None, {'col1': StandardScaler(), 'col2': StandardScaler()})], 
    ids=lambda params: f'params: {params}')
def test_initialize_transformer_series_when_transformer_series_is_StandardScaler(forecaster_name, encoding, expected):
    """
    Test initialize_transformer_series when `transformer_series` is a StandardScaler.
    """
    series_names_in_ = ['col1', 'col2']
    transformer_series = StandardScaler()
    
    transformer_series_ = initialize_transformer_series(
                              forecaster_name    = forecaster_name,
                              series_names_in_   = series_names_in_,
                              encoding           = encoding,
                              transformer_series = transformer_series
                          )

    assert transformer_series_.keys() == expected.keys()
    for k in expected.keys():
        assert isinstance(transformer_series_[k], StandardScaler)


@pytest.mark.parametrize(
    "forecaster_name, encoding, expected", 
    [('ForecasterRecursiveMultiSeries', 'onehot', {'col1': StandardScaler(), 'col2': StandardScaler(), '_unknown_level': StandardScaler()}),
     ('ForecasterDirectMultiVariate', None, {'col1': StandardScaler(), 'col2': StandardScaler()}),
     ('ForecasterRNN', None, {'col1': StandardScaler(), 'col2': StandardScaler()})], 
    ids=lambda params: f'params: {params}')
def test_initialize_transformer_series_when_transformer_series_is_dict(forecaster_name, encoding, expected):
    """
    Test initialize_transformer_series when `transformer_series` is a dict.
    Encoding can not be None when transformer_series is a dict.
    """
    series_names_in_ = ['col1', 'col2']
    if forecaster_name == 'ForecasterRecursiveMultiSeries':
        transformer_series = {'col1': StandardScaler(), 
                              'col2': StandardScaler(), 
                              '_unknown_level': StandardScaler()}
    else:
        transformer_series = {'col1': StandardScaler(), 
                              'col2': StandardScaler()}
    
    transformer_series_ = initialize_transformer_series(
                              forecaster_name    = forecaster_name,
                              series_names_in_   = series_names_in_,
                              encoding           = encoding,
                              transformer_series = transformer_series
                          )

    assert transformer_series_.keys() == expected.keys()
    for k in expected.keys():
        assert isinstance(transformer_series_[k], StandardScaler)


@pytest.mark.parametrize(
    "forecaster_name, encoding, expected", 
    [('ForecasterRecursiveMultiSeries', 'onehot', {'col1': StandardScaler(), 'col2': None, '_unknown_level': StandardScaler()}),
     ('ForecasterDirectMultiVariate', None, {'col1': StandardScaler(), 'col2': None}),
     ('ForecasterRNN', None, {'col1': StandardScaler(), 'col2': None})], 
    ids=lambda params: f'params: {params}')
def test_initialize_transformer_series_IgnoredArgumentWarning_when_levels_of_transformer_series_not_equal_to_series_names_in_(forecaster_name, encoding, expected):
    """
    Test IgnoredArgumentWarning is raised when `transformer_series` is a dict and its keys 
    are not the same as series_names_in_.
    """
    series_names_in_ = ['col1', 'col2']
    if forecaster_name == 'ForecasterRecursiveMultiSeries':
        transformer_series = {'col1': StandardScaler(), 
                              'col3': StandardScaler(), 
                              '_unknown_level': StandardScaler()}
    else:
        transformer_series = {'col1': StandardScaler(), 
                              'col3': StandardScaler()}
    
    series_not_in_transformer_series = set(['col2'])
    warn_msg = re.escape(
        (f"{series_not_in_transformer_series} not present in `transformer_series`."
         f" No transformation is applied to these series.")
    )
    with pytest.warns(IgnoredArgumentWarning, match = warn_msg):
        transformer_series_ = initialize_transformer_series(
                                  forecaster_name    = forecaster_name,
                                  series_names_in_   = series_names_in_,
                                  encoding           = encoding,
                                  transformer_series = transformer_series
                              )

    assert transformer_series_.keys() == expected.keys()
    for k in expected.keys():
        if expected[k] is None:
            assert isinstance(transformer_series_[k], type(None))
        else:
            assert isinstance(transformer_series_[k], StandardScaler)


# Unit test _binning_in_sample_residuals ForecasterRecursiveMultiseries
# ==============================================================================
from pytest import approx
import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression
from skforecast.recursive import ForecasterRecursiveMultiSeries
from skforecast.preprocessing import reshape_series_wide_to_long


def test_binning_in_sample_residuals_output():
    """
    Test that _binning_in_sample_residuals returns the expected output.
    """

    forecaster = ForecasterRecursiveMultiSeries(
        regressor=LinearRegression(), lags=5, binner_kwargs={"n_bins": 3}
    )

    series = pd.DataFrame(
        {
            "l1": pd.Series(np.arange(10)),
            "l2": pd.Series(np.arange(10)),
            "l3": pd.Series(np.arange(10)),
        }
    )
    series.index = pd.date_range(start="2022-01-04", periods=10, freq="D")
    series = reshape_series_wide_to_long(series)

    forecaster.fit(series=series)
    forecaster.in_sample_residuals_ = {}
    forecaster.in_sample_residuals_by_bin_ = {}
    forecaster.binner_intervals_ = {}

    rng = np.random.default_rng(12345)
    y_pred = rng.normal(100, 15, 20)
    y_true = rng.normal(100, 10, 20)

    forecaster._binning_in_sample_residuals(
        level="level_1",
        y_pred=y_pred,
        y_true=y_true,
        store_in_sample_residuals=True
    )

    expected_residuals = {
        "level_1": np.array(
            [
                34.58035615,
                -21.95291202,
                22.08911948,
                -12.32822882,
                -0.451743,
                15.6081091,
                7.0808798,
                -10.55026794,
                11.83152763,
                55.47454021,
                -27.43753138,
                -6.24112163,
                1.80092458,
                -25.62685698,
                -7.11862253,
                6.32581108,
                -4.31327121,
                12.2624188,
                -20.92461257,
                -18.40910724,
            ]
        )
    }

    expected_residuals_by_bin = {
        "level_1": {
            0: np.array(
                [
                    34.58035615,
                    22.08911948,
                    15.6081091,
                    7.0808798,
                    55.47454021,
                    1.80092458,
                    12.2624188,
                ]
            ),
            1: np.array(
                [
                    -12.32822882,
                    -0.45174300,
                    11.83152763,
                    -7.11862253,
                    6.32581108,
                    -20.92461257,
                ]
            ),
            2: np.array(
                [
                    -21.95291202,
                    -10.55026794,
                    -27.43753138,
                    -6.24112163,
                    -25.62685698,
                    -4.31327121,
                    -18.40910724,
                ]
            ),
        }
    }

    expected_binner_intervals = {
        "level_1": {
            0: (70.70705405481715, 90.25638761254116),
            1: (90.25638761254116, 109.36821559391004),
            2: (109.36821559391004, 135.2111448156828),
        },
    }

    assert all(
        np.allclose(forecaster.in_sample_residuals_[key], expected_residuals[key])
        for key in forecaster.in_sample_residuals_
    )
    for k in expected_residuals_by_bin.keys():
        assert forecaster.in_sample_residuals_by_bin_[k].keys() == expected_residuals_by_bin[k].keys()
        for bin in forecaster.in_sample_residuals_by_bin_[k].keys():
            np.testing.assert_almost_equal(
                forecaster.in_sample_residuals_by_bin_[k][bin],
                expected_residuals_by_bin[k][bin],
            )
    assert forecaster.binner_intervals_ == expected_binner_intervals


def test_binning_in_sample_residuals_store_in_sample_residuals_False():
    """
    Test that _binning_in_sample_residuals store_in_sample_residuals False.
    """

    forecaster = ForecasterRecursiveMultiSeries(
        regressor=LinearRegression(), lags=5, binner_kwargs={"n_bins": 3}
    )

    series = pd.DataFrame(
        {
            "l1": pd.Series(np.arange(10)),
            "l2": pd.Series(np.arange(10)),
            "l3": pd.Series(np.arange(10)),
        }
    )
    series.index = pd.date_range(start="2022-01-04", periods=10, freq="D")
    series = reshape_series_wide_to_long(series)

    forecaster.fit(series=series, store_in_sample_residuals=False)
    forecaster.in_sample_residuals_ = None
    forecaster.in_sample_residuals_by_bin_ = None
    forecaster.binner_intervals_ = {}

    rng = np.random.default_rng(12345)
    y_pred = rng.normal(100, 15, 20)
    y_true = rng.normal(100, 10, 20)

    forecaster._binning_in_sample_residuals(
        level="level_1",
        y_pred=y_pred,
        y_true=y_true,
        store_in_sample_residuals=False
    )
    results_residuals = forecaster.in_sample_residuals_
    results_residuals_bin = forecaster.in_sample_residuals_by_bin_
    results_binner_intervals = forecaster.binner_intervals_

    expected_residuals = None
    expected_residuals_by_bin = None
    expected_binner_intervals = {
        "level_1": {
            0: (70.70705405481715, 90.25638761254116),
            1: (90.25638761254116, 109.36821559391004),
            2: (109.36821559391004, 135.2111448156828),
        },
    }

    # In-sample residuals
    assert results_residuals == expected_residuals

    # In-sample residuals by bin
    assert results_residuals_bin == expected_residuals_by_bin

    # Binner intervals
    assert results_binner_intervals.keys() == expected_binner_intervals.keys()
    for level in results_binner_intervals.keys():
        assert results_binner_intervals[level].keys() == expected_binner_intervals[level].keys()
        for k in results_binner_intervals[level].keys():
            assert results_binner_intervals[level][k][0] == approx(expected_binner_intervals[level][k][0])
            assert results_binner_intervals[level][k][1] == approx(expected_binner_intervals[level][k][1])


def test_binning_in_sample_residuals_probabilistic_mode_no_binned():
    """
    Test that _binning_in_sample_residuals stores when _probabilistic_mode is
    no binned.
    """

    forecaster = ForecasterRecursiveMultiSeries(
        regressor=LinearRegression(), lags=5, binner_kwargs={"n_bins": 3}
    )

    series = pd.DataFrame(
        {
            "l1": pd.Series(np.arange(10)),
            "l2": pd.Series(np.arange(10)),
            "l3": pd.Series(np.arange(10)),
        }
    )
    series.index = pd.date_range(start="2022-01-04", periods=10, freq="D")
    series = reshape_series_wide_to_long(series)

    forecaster.fit(series=series)
    forecaster.in_sample_residuals_ = {}
    forecaster.in_sample_residuals_by_bin_ = {}
    forecaster.binner_intervals_ = {}

    rng = np.random.default_rng(12345)
    y_pred = rng.normal(100, 15, 20)
    y_true = rng.normal(100, 10, 20)

    forecaster._probabilistic_mode = "no_binned"
    forecaster._binning_in_sample_residuals(
        level="level_1",
        y_pred=y_pred,
        y_true=y_true,
        store_in_sample_residuals=True
    )
    results_residuals = forecaster.in_sample_residuals_

    expected_residuals = {
        "level_1": np.array(
            [
                34.58035615,
                -21.95291202,
                22.08911948,
                -12.32822882,
                -0.451743,
                15.6081091,
                7.0808798,
                -10.55026794,
                11.83152763,
                55.47454021,
                -27.43753138,
                -6.24112163,
                1.80092458,
                -25.62685698,
                -7.11862253,
                6.32581108,
                -4.31327121,
                12.2624188,
                -20.92461257,
                -18.40910724,
            ]
        )
    }

    # In-sample residuals
    assert isinstance(results_residuals, dict)
    assert results_residuals.keys() == expected_residuals.keys()
    assert np.all(results_residuals[k] == expected_residuals[k] for k in results_residuals.keys())

    assert forecaster.in_sample_residuals_by_bin_ == {}
    assert forecaster.binner_intervals_ == {}


def test_binning_in_sample_residuals_stores_maximum_10000_residuals_per_level():
    """
    Test that maximum 10_000 residuals are stored per ñevel and 10_000 / n_bins per bin.
    """

    forecaster = ForecasterRecursiveMultiSeries(
        regressor=LinearRegression(), lags=5, binner_kwargs={"n_bins": 3}
    )

    rng = np.random.default_rng(12345)
    series = pd.DataFrame(
        {
            "l1": rng.normal(100, 15, 15_000),
            "l2": rng.normal(100, 15, 15_000),
            "l3": rng.normal(100, 15, 15_000),
        },
        index=pd.date_range(start="1990-01-01", periods=15_000, freq="h"),
    )
    series = reshape_series_wide_to_long(series)
    
    forecaster.fit(series=series)
    forecaster.in_sample_residuals_ = {}
    forecaster.in_sample_residuals_by_bin_ = {}
    
    y_pred = rng.normal(100, 15, 15_000)
    y_true = rng.normal(100, 10, 15_000)

    forecaster._binning_in_sample_residuals(
        level="level_1",
        y_pred=y_pred,
        y_true=y_true,
        store_in_sample_residuals=True
    )

    for k in forecaster.in_sample_residuals_.keys():
        assert len(forecaster.in_sample_residuals_[k]) == 10_000

    for k in forecaster.in_sample_residuals_by_bin_.keys():
        max_size = 10_000 // 3
        for bin in [0, 1, 2]:
            assert len(forecaster.in_sample_residuals_by_bin_[k][bin]) == max_size



# Unit test predict_quantiles ForecasterRecursive
# ==============================================================================
import numpy as np
import pandas as pd
from skforecast.recursive import ForecasterRecursive
from sklearn.linear_model import LinearRegression
from sklearn.preprocessing import StandardScaler

# Fixtures
from .fixtures_forecaster_recursive import y
from .fixtures_forecaster_recursive import exog
from .fixtures_forecaster_recursive import exog_predict


def test_predict_quantiles_output_when_forecaster_is_LinearRegression_steps_is_2_in_sample_residuals_True_exog_and_transformer():
    """
    Test output of predict_quantiles when regressor is LinearRegression,
    2 steps are predicted, using in-sample residuals, exog is included and both
    inputs are transformed.
    """
    forecaster = ForecasterRecursive(
                     regressor        = LinearRegression(),
                     lags             = 3,
                     transformer_y    = StandardScaler(),
                     transformer_exog = StandardScaler()
                 )

    forecaster.fit(y=y, exog=exog, store_in_sample_residuals=True)
    results = forecaster.predict_quantiles(
                  steps                   = 2,
                  exog                    = exog_predict,
                  quantiles               = [0.05, 0.55, 0.95],
                  n_boot                  = 4,
                  use_in_sample_residuals = True,
                  use_binned_residuals    = False
              )
    
    expected = pd.DataFrame(
                   data    = np.array([[0.25836446, 0.571056  , 0.71651764],
                                       [0.10646288, 0.32036708, 0.64980398]]),
                   columns = ['q_0.05', 'q_0.55', 'q_0.95'],
                   index   = pd.RangeIndex(start=50, stop=52)
               )
    
    pd.testing.assert_frame_equal(expected, results)


def test_predict_quantiles_output_when_forecaster_is_LinearRegression_steps_is_2_in_sample_residuals_False_exog_and_transformer():
    """
    Test output of predict_quantiles when regressor is LinearRegression,
    2 steps are predicted, using out-sample residuals, exog is included and both
    inputs are transformed.
    """
    forecaster = ForecasterRecursive(
                     regressor        = LinearRegression(),
                     lags             = 3,
                     transformer_y    = StandardScaler(),
                     transformer_exog = StandardScaler(),
                     binner_kwargs    = {'n_bins': 15}
                 )
    
    forecaster.fit(y=y, exog=exog, store_in_sample_residuals=True)
    forecaster.out_sample_residuals_ = forecaster.in_sample_residuals_
    results = forecaster.predict_quantiles(
                  steps                   = 2,
                  exog                    = exog_predict,
                  quantiles               = (0.05, 0.55, 0.95),
                  n_boot                  = 4,
                  use_in_sample_residuals = False,
                  use_binned_residuals    = False
              )
    
    expected = pd.DataFrame(
                   data    = np.array([[0.25836446, 0.571056  , 0.71651764],
                                       [0.10646288, 0.32036708, 0.64980398]]),
                   columns = ['q_0.05', 'q_0.55', 'q_0.95'],
                   index   = pd.RangeIndex(start=50, stop=52)
               )

    pd.testing.assert_frame_equal(expected, results)



# Unit test random_search_forecaster
# ==============================================================================
import numpy as np
import pandas as pd
from sklearn.linear_model import Ridge
from skforecast.recursive import ForecasterRecursive
from skforecast.model_selection._search import random_search_forecaster
from skforecast.model_selection._split import TimeSeriesFold

# Fixtures
from ..fixtures_model_selection import y

from tqdm import tqdm
from functools import partialmethod
tqdm.__init__ = partialmethod(tqdm.__init__, disable=True)  # hide progress bar


def test_output_random_search_forecaster_ForecasterRecursive_with_mocked():
    """
    Test output of random_search_forecaster in ForecasterRecursive with mocked
    (mocked done in Skforecast v0.4.3)
    """
    forecaster = ForecasterRecursive(
                    regressor = Ridge(random_state=123),
                    lags      = 2
                 )
    n_validation = 12
    y_train = y[:-n_validation]
    lags_grid = [2, 4]
    param_distributions = {'alpha': np.logspace(-5, 3, 10)}
    n_iter = 3
    cv = TimeSeriesFold(
            steps                 = 3,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = False,
            fixed_train_size      = False,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False
         )

    results = random_search_forecaster(
                  forecaster           = forecaster,
                  y                    = y,
                  cv                   = cv,
                  lags_grid            = lags_grid,
                  param_distributions  = param_distributions,
                  metric               = 'mean_squared_error',
                  n_iter               = n_iter,
                  random_state         = 123,
                  return_best          = False,
                  verbose              = False
              )

    expected_results = pd.DataFrame(
        {
            "lags": [[1, 2], [1, 2], [1, 2, 3, 4], [1, 2, 3, 4], [1, 2, 3, 4], [1, 2]],
            "lags_label": [
                [1, 2],
                [1, 2],
                [1, 2, 3, 4],
                [1, 2, 3, 4],
                [1, 2, 3, 4],
                [1, 2],
            ],
            "params": [
                {"alpha": 1e-05},
                {"alpha": 0.03593813663804626},
                {"alpha": 1e-05},
                {"alpha": 0.03593813663804626},
                {"alpha": 16.681005372000556},
                {"alpha": 16.681005372000556},
            ],
            "mean_squared_error": np.array(
                [0.06460234, 0.06475887, 0.06776596, 0.06786132, 0.0713478, 0.07161]
            ),
            "alpha": np.array(
                [
                    1.00000000e-05,
                    3.59381366e-02,
                    1.00000000e-05,
                    3.59381366e-02,
                    1.66810054e01,
                    1.66810054e01,
                ]
            ),
        },
        index=pd.Index([0, 1, 2, 3, 4, 5], dtype="int64"),
    )

    pd.testing.assert_frame_equal(results, expected_results)



# Unit test check_extract_values_and_index
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from skforecast.utils import check_extract_values_and_index


def test_check_extract_values_and_index_ValueError_when_DatetimeIndex_without_frequency():
    """
    Check ValueError is raised when the index of data is a pandas DatetimeIndex
    without a frequency.
    """

    data = pd.Series([1, 2, 3], name='exog') 
    data.index = pd.date_range("2020-01-01", periods=3, freq=None)
    data.index.freq = None  # Remove frequency to simulate the error

    err_msg = re.escape(
        "`exog` has a pandas DatetimeIndex without a frequency. "
        "To avoid this error, set the frequency of the DatetimeIndex."
    )
    with pytest.raises(ValueError, match = err_msg):
        check_extract_values_and_index(data=data, data_label='`exog`')


def test_check_extract_values_and_index_TypeError_when_no_DatetimeIndex_or_RangeIndex():
    """
    Check TypeError is raised when the index of data is neither a pandas
    DatetimeIndex nor a RangeIndex.
    """

    data = pd.Series([1, 2, 3], name='last_window') 
    data.index = pd.Index(['a', 'b', 'c'])  # Non-DatetimeIndex or RangeIndex

    err_msg = re.escape(
        f"`last_window` has an unsupported index type. The index must be a "
        f"pandas DatetimeIndex or a RangeIndex. Got {type(data.index)}."
    )
    with pytest.raises(TypeError, match = err_msg):
        check_extract_values_and_index(data=data, data_label='`last_window`')


@pytest.mark.parametrize("idx", 
                         [pd.date_range("2020-01-01", periods=3, freq='D'),
                          pd.RangeIndex(start=0, stop=3)], 
                         ids = lambda idx: f'index type: {type(idx)}')
def test_check_extract_values_and_index_output(idx):
    """
    Test check_extract_values_and_index returns the correct values and index
    when the index is a pandas DatetimeIndex or a RangeIndex.
    """
    data = pd.Series([1, 2, 3], name='exog', index=idx)
    data_values, data_index = check_extract_values_and_index(data=data, data_label='`exog`')
    
    np.testing.assert_array_almost_equal(data_values, np.array([1, 2, 3]))
    pd.testing.assert_index_equal(data_index, idx)



# Unit test initialize_weights
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
import inspect
from skforecast.exceptions import IgnoredArgumentWarning
from skforecast.utils.utils import initialize_weights
from sklearn.linear_model import LinearRegression
from sklearn.ensemble import RandomForestRegressor
from sklearn.neighbors import KNeighborsRegressor


@pytest.mark.parametrize("forecaster_name", 
                         ['ForecasterRecursive', 'ForecasterDirect', 'ForecasterDirectMultiVariate'], 
                         ids=lambda fn: f'forecaster_name: {fn}')
def test_TypeError_initialize_weights_when_weight_func_is_not_a_Callable(forecaster_name):
    """
    Test TypeError is raised when weight_func is not a Callable.
    """
    weight_func = 'not_Callable'
    err_msg = re.escape(
        f"Argument `weight_func` must be a Callable. Got {type(weight_func)}."
    )
    with pytest.raises(TypeError, match = err_msg):
        initialize_weights(
            forecaster_name = forecaster_name, 
            regressor       = LinearRegression(), 
            weight_func     = weight_func, 
            series_weights  = None
        )


def test_TypeError_initialize_weights_when_weight_func_is_not_a_Callable_or_dict():
    """
    Test TypeError is raised when weight_func is not a Callable or a dict.
    """
    weight_func = 'not_Callable_or_dict'
    err_msg = re.escape(
        f"Argument `weight_func` must be a Callable or a dict of Callables. Got {type(weight_func)}."
    )
    with pytest.raises(TypeError, match = err_msg):
        initialize_weights(
            forecaster_name = 'ForecasterRecursiveMultiSeries', 
            regressor       = LinearRegression(), 
            weight_func     = weight_func, 
            series_weights  = None
        )


def test_TypeError_initialize_weights_when_series_weights_is_not_a_dict():
    """
    Test TypeError is raised when series_weights is not a dict.
    """
    series_weights = 'not_Callable_or_dict'
    err_msg = re.escape(
        (f"Argument `series_weights` must be a dict of floats or ints."
         f"Got {type(series_weights)}.")
    )
    with pytest.raises(TypeError, match = err_msg):
        initialize_weights(
            forecaster_name = 'ForecasterRecursiveMultiSeries', 
            regressor       = LinearRegression(), 
            weight_func     = None, 
            series_weights  = series_weights
        )


def test_IgnoredArgumentWarning_initialize_weights_when_weight_func_is_provided_and_regressor_has_not_sample_weights():
    """
    Test IgnoredArgumentWarning is created when weight_func is provided but the regressor 
    has no argument sample_weights in his fit method.
    """
    def weight_func():  # pragma: no cover
        pass

    warn_msg = re.escape(
        ("Argument `weight_func` is ignored since regressor KNeighborsRegressor() "
         "does not accept `sample_weight` in its `fit` method.")
    )
    with pytest.warns(IgnoredArgumentWarning, match = warn_msg):
        weight_func, source_code_weight_func, _ = initialize_weights(
            forecaster_name = 'ForecasterRecursive', 
            regressor       = KNeighborsRegressor(), 
            weight_func     = weight_func, 
            series_weights  = None
        )
    
    assert weight_func is None
    assert source_code_weight_func is None


def test_IgnoredArgumentWarning_initialize_weights_when_series_weights_is_provided_and_regressor_has_not_sample_weights():
    """
    Test IgnoredArgumentWarning is created when series_weights is provided but the regressor 
    has no argument sample_weights in his fit method.
    """
    series_weights = {'series_1': 1., 'series_2': 2.}

    warn_msg = re.escape(
                ("Argument `series_weights` is ignored since regressor KNeighborsRegressor() "
                 "does not accept `sample_weight` in its `fit` method.")
            )
    with pytest.warns(IgnoredArgumentWarning, match = warn_msg):
        weight_func, source_code_weight_func, series_weights = initialize_weights(
            forecaster_name = 'ForecasterRecursiveMultiSeries', 
            regressor       = KNeighborsRegressor(), 
            weight_func     = None, 
            series_weights  = series_weights
        )
    
    assert series_weights is None


@pytest.mark.parametrize("regressor", 
                         [LinearRegression(), RandomForestRegressor()], 
                         ids=lambda regressor: f'regressor: {type(regressor).__name__}')
def test_initialize_weights_finds_sample_weight_in_different_regressors_when_weight_func(recwarn, regressor):
    """
    Test initialize weights finds `sample_weight` attribute in different
    regressors when `weight_func`.
    """
    def weight_func():  # pragma: no cover
        pass

    weight_func, source_code_weight_func, _ = initialize_weights(
        forecaster_name = 'ForecasterRecursive', 
        regressor       = regressor, 
        weight_func     = weight_func, 
        series_weights  = None
    )
    
    # Count the number of warnings, it should be 0
    assert len(recwarn) == 0


@pytest.mark.parametrize("regressor", 
                         [LinearRegression(), RandomForestRegressor()], 
                         ids=lambda regressor: f'regressor: {type(regressor).__name__}')
def test_initialize_weights_finds_sample_weight_in_different_regressors_when_series_weights(recwarn, regressor):
    """
    Test initialize weights finds `sample_weight` attribute in different
    regressors when `series_weights`.
    """
    series_weights = {'series_1': 1., 'series_2': 2.}

    weight_func, source_code_weight_func, series_weights = initialize_weights(
        forecaster_name = 'ForecasterRecursiveMultiSeries', 
        regressor       = regressor, 
        weight_func     = None, 
        series_weights  = series_weights
    )
    
    # Count the number of warnings, it should be 0
    assert len(recwarn) == 0


def test_output_initialize_weights_source_code_weight_func_when_weight_func_not_dict():
    """
    Test source_code_weight_func output of initialize_weights when 
    weight_func is a Callable.
    """
    def test_weight_func():  # pragma: no cover
        """
        test
        """
        pass

    weight_func, source_code_weight_func, series_weights = initialize_weights(
        forecaster_name = 'ForecasterRecursiveMultiSeries', 
        regressor       = LinearRegression(), 
        weight_func     = test_weight_func, 
        series_weights  = None
    )
    
    assert source_code_weight_func == inspect.getsource(test_weight_func)


def test_output_initialize_weights_source_code_weight_func_when_weight_func_dict():
    """
    Test source_code_weight_func output of initialize_weights when 
    weight_func is a dict.
    """
    def test_weight_func():  # pragma: no cover
        """
        test
        """
        pass

    def test_weight_func_2():  # pragma: no cover
        """
        test2
        """
        pass

    weight_func = {'series_1': test_weight_func, 'series_2': test_weight_func_2}    

    weight_func, source_code_weight_func, series_weights = initialize_weights(
        forecaster_name = 'ForecasterRecursiveMultiSeries', 
        regressor       = LinearRegression(), 
        weight_func     = weight_func, 
        series_weights  = None
    )
    
    assert source_code_weight_func['series_1'] == inspect.getsource(test_weight_func)
    assert source_code_weight_func['series_2'] == inspect.getsource(test_weight_func_2)



################################################################################
#                                ForecasterBase                                #
#                                                                              #
# This work by skforecast team is licensed under the BSD 3-Clause License.     #
################################################################################
# coding=utf-8

from __future__ import annotations
from abc import ABC, abstractmethod
import textwrap
import numpy as np
import pandas as pd
from sklearn.pipeline import Pipeline


class ForecasterBase(ABC):
    """
    Base class for all forecasters in skforecast. All forecasters should specify
    all the parameters that can be set at the class level in their ``__init__``.     
    """

    def _preprocess_repr(
        self,
        regressor: object | None = None,
        training_range_: dict[str, str] | None = None,
        series_names_in_: list[str] | None = None,
        exog_names_in_: list[str] | None = None,
        transformer_series: object | dict[str, object] | None = None
    ) -> tuple[str, str | None, str | None, str | None, str | None]:
        """
        Prepare the information to be displayed when a Forecaster object is printed.

        Parameters
        ----------
        regressor : object, default None
            Regressor object.
        training_range_ : dict, default None
            Training range. Only used for `ForecasterRecursiveMultiSeries`.
        series_names_in_ : list, default None
            Names of the series used in the forecaster. Only used for `ForecasterRecursiveMultiSeries`.
        exog_names_in_ : list, default None
            Names of the exogenous variables used in the forecaster.
        transformer_series : object, dict, default None
            Transformer used in the series. Only used for `ForecasterRecursiveMultiSeries`.

        Returns
        -------
        params : str, None
            Parameters of the regressor.
        training_range_ : str, None
            Training range. Only used for `ForecasterRecursiveMultiSeries`.
        series_names_in_ : str, None
            Names of the series used in the forecaster. Only used for multi-series forecasters.
        exog_names_in_ : str, None
            Names of the exogenous variables used in the forecaster.
        transformer_series : str, None
            Transformer used in the series. Only used for multi-series forecasters.
        
        """

        if regressor is not None:
            if isinstance(regressor, Pipeline):
                name_pipe_steps = tuple(name + "__" for name in regressor.named_steps.keys())
                params = {
                    key: value for key, value in regressor.get_params().items() 
                    if key.startswith(name_pipe_steps)
                }
            else:
                params = regressor.get_params()
            params = str(params)
        else:
            params = None

        if training_range_ is not None:
            training_range_ = [
                f"'{k}': {v.astype(str).to_list()}" 
                for k, v in training_range_.items()
            ]
            if len(training_range_) > 10:
                training_range_ = training_range_[:5] + ['...'] + training_range_[-5:]
            training_range_ = ", ".join(training_range_)

        if series_names_in_ is not None:
            if len(series_names_in_) > 50:
                series_names_in_ = series_names_in_[:25] + ["..."] + series_names_in_[-25:]
            series_names_in_ = ", ".join(series_names_in_)

        if exog_names_in_ is not None:
            if len(exog_names_in_) > 50:
                exog_names_in_ = exog_names_in_[:25] + ["..."] + exog_names_in_[-25:]
            exog_names_in_ = ", ".join(exog_names_in_)
        
        if transformer_series is not None:
            if isinstance(transformer_series, dict):
                transformer_series = [
                    f"'{k}': {v}" 
                    for k, v in transformer_series.items()
                ]
                if len(transformer_series) > 10:
                    transformer_series = transformer_series[:5] + ["..."] + transformer_series[-5:]
                transformer_series = ", ".join(transformer_series)
            else:
                transformer_series = str(transformer_series)        

        return params, training_range_, series_names_in_, exog_names_in_, transformer_series

    def _format_text_repr(
        self, 
        text: str, 
        max_text_length: int = 58,
        width: int = 80, 
        indent: str = "    "
    ) -> str:
        """
        Format text for __repr__ method.

        Parameters
        ----------
        text : str
            Text to format.
        max_text_length : int, default 58
            Maximum length of the text before wrapping.
        width : int, default 80
            Maximum width of the text.
        indent : str, default "    "
            Indentation of the text.
        
        Returns
        -------
        text : str
            Formatted text.

        """

        if text is not None and len(text) > max_text_length:
            text = "\n    " + textwrap.fill(
                str(text), width=width, subsequent_indent=indent
            )
        
        return text

    @abstractmethod
    def create_train_X_y(
        self,
        y: pd.Series,
        exog: pd.Series | pd.DataFrame | None = None
    ) -> tuple[pd.DataFrame, pd.Series]:
        """
        Create training matrices from univariate time series and exogenous
        variables.

        Parameters
        ----------
        y : pandas Series
            Training time series.
        exog : pandas Series, pandas DataFrame, default None
            Exogenous variable/s included as predictor/s. Must have the same
            number of observations as `y` and their indexes must be aligned.

        Returns
        -------
        X_train : pandas DataFrame
            Training values (predictors).
            Shape: (len(y) - self.max_lag, len(self.lags))
        y_train : pandas Series
            Values (target) of the time series related to each row of `X_train`.
            Shape: (len(y) - self.max_lag, )
        
        """
        
        pass

    @abstractmethod
    def fit(
        self,
        y: pd.Series,
        exog: pd.Series | pd.DataFrame | None = None
    ) -> None:
        """
        Training Forecaster.
        
        Parameters
        ----------
        y : pandas Series
            Training time series.
        exog : pandas Series, pandas DataFrame, default None
            Exogenous variable/s included as predictor/s. Must have the same
            number of observations as `y` and their indexes must be aligned so
            that y[i] is regressed on exog[i].

        Returns
        -------
        None
        
        """
        
        pass

    @abstractmethod        
    def predict(
        self,
        steps: int,
        last_window: pd.Series | pd.DataFrame | None = None,
        exog: pd.Series | pd.DataFrame | None = None
    ) -> pd.Series:
        """
        Predict n steps ahead.
        
        Parameters
        ----------
        steps : int
            Number of steps to predict. 
        last_window : pandas Series, pandas DataFrame, default None
            Series values used to create the predictors (lags) needed in the 
            first iteration of the prediction (t + 1).
            If `last_window = None`, the values stored in `self.last_window` are
            used to calculate the initial predictors, and the predictions start
            right after training data.
        exog : pandas Series, pandas DataFrame, default None
            Exogenous variable/s included as predictor/s.

        Returns
        -------
        predictions : pandas Series
            Predicted values.
        
        """

        pass
        
    @abstractmethod
    def set_params(self, params: dict[str, object]) -> None:
        """
        Set new values to the parameters of the scikit-learn model stored in the
        forecaster.
        
        Parameters
        ----------
        params : dict
            Parameters values.

        Returns
        -------
        None
        
        """
        
        pass
        
    def set_lags(
        self, 
        lags: int | list[int] | np.ndarray[int] | range[int] | None = None
    ) -> None:
        """
        Set new value to the attribute `lags`.
        Attributes `max_lag` and `window_size` are also updated.
        
        Parameters
        ----------
        lags : int, list, numpy ndarray, range, default None
            Lags used as predictors. Index starts at 1, so lag 1 is equal to t-1. 
        
            - `int`: include lags from 1 to `lags` (included).
            - `list`, `1d numpy ndarray` or `range`: include only lags present in 
            `lags`, all elements must be int.
            - `None`: no lags are included as predictors. 

        Returns
        -------
        None
        
        """
        
        pass

    def set_window_features(
        self, 
        window_features: object | list[object] | None = None
    ) -> None:
        """
        Set new value to the attribute `window_features`. Attributes 
        `max_size_window_features`, `window_features_names`, 
        `window_features_class_names` and `window_size` are also updated.
        
        Parameters
        ----------
        window_features : object, list, default None
            Instance or list of instances used to create window features. Window features
            are created from the original time series and are included as predictors.

        Returns
        -------
        None
        
        """

        pass

    def summary(self) -> None:
        """
        Show forecaster information.
        
        Parameters
        ----------
        self

        Returns
        -------
        None
        
        """
        
        print(self.__repr__())



# Unit test check_select_fit_kwargs
# ==============================================================================
import re
import pytest
import inspect
from skforecast.exceptions import IgnoredArgumentWarning
from skforecast.utils import check_select_fit_kwargs
from lightgbm import LGBMRegressor


def test_check_select_fit_kwargs_raise_TypeError_when_fit_kwargs_is_not_dict():
    """
    Test check_select_fit_kwargs raises TypeError when fit_kwargs is not a dictionary.
    """
    fit_kwargs = 'not_dict'
    regressor = LGBMRegressor()

    err_msg = re.escape(
            f"Argument `fit_kwargs` must be a dict. Got {type(fit_kwargs)}."
        )
    with pytest.raises(TypeError, match = err_msg):
        check_select_fit_kwargs(regressor=regressor, fit_kwargs=fit_kwargs)



def test_check_select_fit_kwargs_IgnoredArgumentWarning_when_fit_kwargs_has_arguments_not_in_fit():
    """
    Test check_select_fit_kwargs issues IgnoredArgumentWarning when fit_kwargs 
    has arguments not in regressor fit method.
    """
    fit_kwargs = {'no_valid_argument': 10}
    regressor = LGBMRegressor()
    non_used_keys = [k for k in fit_kwargs.keys()
                     if k not in inspect.signature(regressor.fit).parameters]
    
    warn_msg = re.escape(
            (f"Argument/s {non_used_keys} ignored since they are not used by the "
             f"regressor's `fit` method.")
        )
    with pytest.warns(IgnoredArgumentWarning, match = warn_msg):       
        check_select_fit_kwargs(regressor=regressor, fit_kwargs=fit_kwargs)


def test_check_select_fit_kwargs_IgnoredArgumentWarning_when_fit_kwargs_has_sample_weight():
    """
    Test check_select_fit_kwargs issues IgnoredArgumentWarning when fit_kwargs 
    has sample_weight argument.
    """
    fit_kwargs = {'sample_weight': [1,2,3]}
    regressor = LGBMRegressor()
    
    warn_msg = re.escape(
            ("The `sample_weight` argument is ignored. Use `weight_func` to pass "
             "a function that defines the individual weights for each sample "
             "based on its index.")
        )
    with pytest.warns(IgnoredArgumentWarning, match = warn_msg):       
        check_select_fit_kwargs(regressor=regressor, fit_kwargs=fit_kwargs)


def test_check_select_fit_kwargs_ignores_sample_weight():
    """
    Test check_select_fit_kwargs method ignores sample_weight argument.
    """
    results = check_select_fit_kwargs(
                  regressor  = LGBMRegressor(),
                  fit_kwargs = {'sample_weight': [1,2,3]}
              )
    expected = {}

    assert results == expected


def test_check_select_fit_kwargs_ignores_arguments_not_in_regressor_fit():
    """
    Test check_select_fit_kwargs method ignores arguments not in regressor fit method.
    """
    results = check_select_fit_kwargs(
                  regressor  = LGBMRegressor(),
                  fit_kwargs = {'feature_name':'auto', 'no_valid_argument': 10}
              )
    expected = {'feature_name':'auto'}

    assert results == expected


# Unit test _check_X_numpy_ndarray_1d
# ==============================================================================
import re
import numpy as np
import pandas as pd
import pytest
from skforecast.preprocessing import TimeSeriesDifferentiator


def test_TypeError_decorator_check_X_numpy_ndarray_1d_when_X_not_numpy_ndarray():
    """
    Test TypeError raised when X is not a numpy ndarray.
    """
    y = [1, 2, 3, 4, 5]
    differentiator = TimeSeriesDifferentiator(order=1)    

    err_msg = re.escape(
        ("'X' must be a numpy ndarray. Found <class 'list'>.")
    )
    with pytest.raises(TypeError, match = err_msg): 
        differentiator.fit(y)
    with pytest.raises(TypeError, match = err_msg): 
        differentiator.transform(X=y)
    with pytest.raises(TypeError, match = err_msg): 
        differentiator.fit_transform(y)
    with pytest.raises(TypeError, match = err_msg): 
        differentiator.inverse_transform(y)
    with pytest.raises(TypeError, match = err_msg): 
        differentiator.inverse_transform_next_window(y)


def test_ValueError_decorator_check_X_numpy_ndarray_1d_when_X_not_1D_numpy_ndarray():
    """
    Test ValueError raised when X is not a 1D numpy ndarray.
    """
    y = pd.Series(np.array([1, 2, 3, 4, 5])).to_frame().to_numpy()
    differentiator = TimeSeriesDifferentiator(order=1)    

    err_msg = re.escape(
        ("'X' must be a 1D array. Found 2 dimensions.")
    )
    with pytest.raises(ValueError, match = err_msg): 
        differentiator.fit(y)
    with pytest.raises(ValueError, match = err_msg): 
        differentiator.transform(X=y)
    with pytest.raises(ValueError, match = err_msg): 
        differentiator.fit_transform(y)
    with pytest.raises(ValueError, match = err_msg): 
        differentiator.inverse_transform(y)



# Unit test create_predict_X ForecasterRecursive
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from sklearn.exceptions import NotFittedError
from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import StandardScaler
from sklearn.preprocessing import OneHotEncoder
from sklearn.preprocessing import OrdinalEncoder
from sklearn.preprocessing import FunctionTransformer
from sklearn.compose import make_column_transformer, make_column_selector
from sklearn.pipeline import make_pipeline
from sklearn.linear_model import LinearRegression
from sklearn.ensemble import HistGradientBoostingRegressor
from lightgbm import LGBMRegressor

from ....exceptions import DataTransformationWarning
from skforecast.utils import transform_numpy
from skforecast.preprocessing import RollingFeatures
from skforecast.recursive import ForecasterRecursive

# Fixtures
from .fixtures_forecaster_recursive import y as y_categorical
from .fixtures_forecaster_recursive import exog as exog_categorical
from .fixtures_forecaster_recursive import data  # to test results when using differentiation


def test_create_predict_X_NotFittedError_when_fitted_is_False():
    """
    Test NotFittedError is raised when fitted is False.
    """
    forecaster = ForecasterRecursive(
                     regressor = LinearRegression(),
                     lags      = 5
                 )

    err_msg = re.escape(
        "This Forecaster instance is not fitted yet. Call `fit` with "
        "appropriate arguments before using predict."
    )
    with pytest.raises(NotFittedError, match = err_msg):
        forecaster.create_predict_X(steps=5)


def test_create_predict_X_when_regressor_is_LinearRegression():
    """
    Test create_predict_X when using LinearRegression as regressor.
    """
    forecaster = ForecasterRecursive(
                     regressor = LinearRegression(),
                     lags      = 5
                 )
    forecaster.fit(y=pd.Series(np.arange(50, dtype=float), name='y'))
    results = forecaster.create_predict_X(steps=5)

    expected = pd.DataFrame(
        data = {
            'lag_1': [49., 50., 51., 52., 53.],
            'lag_2': [48., 49., 50., 51., 52.],
            'lag_3': [47., 48., 49., 50., 51.],
            'lag_4': [46., 47., 48., 49., 50.],
            'lag_5': [45., 46., 47., 48., 49.]
        },
        index = pd.RangeIndex(start=50, stop=55, step=1)
    )
    
    pd.testing.assert_frame_equal(results, expected)


def test_create_predict_X_when_regressor_is_LinearRegression_and_StandardScaler():
    """
    Test create_predict_X when using LinearRegression as regressor and StandardScaler.
    """
    forecaster = ForecasterRecursive(
                     regressor     = LinearRegression(),
                     lags          = 5,
                     transformer_y = StandardScaler()
                 )
    forecaster.fit(y=pd.Series(np.arange(50, dtype=float), name='y'))
    results = forecaster.create_predict_X(steps=5)

    expected = pd.DataFrame(
        data = np.array([
            [1.69774938, 1.62845348, 1.55915759, 1.4898617 , 1.4205658],
            [1.76704527, 1.69774938, 1.62845348, 1.55915759, 1.4898617],
            [1.83634116, 1.76704527, 1.69774938, 1.62845348, 1.55915759],
            [1.90563705, 1.83634116, 1.76704527, 1.69774938, 1.62845348],
            [1.97493295, 1.90563705, 1.83634116, 1.76704527, 1.69774938]]),
        columns = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5'],
        index = pd.RangeIndex(start=50, stop=55, step=1)
    )
    
    pd.testing.assert_frame_equal(results, expected)


def test_create_predict_X_when_regressor_is_LinearRegression_with_transform_y():
    """
    Test create_predict_X when using LinearRegression as regressor and StandardScaler.
    """
    y = pd.Series(
            np.array([-0.59,  0.02, -0.9 ,  1.09, -3.61,  0.72, -0.11, -0.4 ,  0.49,
                       0.67,  0.54, -0.17,  0.54,  1.49, -2.26, -0.41, -0.64, -0.8 ,
                      -0.61, -0.88]),
            name = 'y'
        )

    forecaster = ForecasterRecursive(
                     regressor     = LinearRegression(),
                     lags          = 5,
                     transformer_y = StandardScaler()
                 )
    forecaster.fit(y=y)
    results = forecaster.create_predict_X(steps=5)

    expected = pd.DataFrame(
        data = np.array([
            [-0.52297655, -0.28324197, -0.45194408, -0.30987914, -0.1056608 ],
            [ 0.11825103, -0.52297655, -0.28324197, -0.45194408, -0.30987914],
            [ 0.0944737 ,  0.11825103, -0.52297655, -0.28324197, -0.45194408],
            [ 0.13663937,  0.0944737 ,  0.11825103, -0.52297655, -0.28324197],
            [ 0.2409212 ,  0.13663937,  0.0944737 ,  0.11825103, -0.52297655]]),
        columns = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5'],
        index = pd.RangeIndex(start=20, stop=25, step=1)
    )
    
    pd.testing.assert_frame_equal(results, expected)


def test_create_predict_X_when_regressor_is_LinearRegression_with_transform_y_and_transform_exog_series():
    """
    Test create_predict_X when using LinearRegression as regressor, StandardScaler
    as transformer_y and StandardScaler as transformer_exog.
    """
    y = pd.Series(np.array([-0.59,  0.02, -0.9 ,  1.09, -3.61,  0.72, -0.11, -0.4]))
    exog = pd.Series(np.array([7.5, 24.4, 60.3, 57.3, 50.7, 41.4, 87.2, 47.4]), name='exog')
    exog_predict = exog.copy()
    exog_predict.index = pd.RangeIndex(start=8, stop=16)

    forecaster = ForecasterRecursive(
                     regressor        = LinearRegression(),
                     lags             = 5,
                     transformer_y    = StandardScaler(),
                     transformer_exog = StandardScaler()
                 )
    forecaster.fit(y=y, exog=exog)
    results = forecaster.create_predict_X(steps=5, exog=exog_predict)

    expected = pd.DataFrame(
        data = np.array([
            [ 0.0542589 ,  0.27129451,  0.89246539, -2.34810076,  1.16937289, -1.76425513],
            [ 0.73245279,  0.0542589 ,  0.27129451,  0.89246539, -2.34810076, -1.00989936],
            [ 0.28154534,  0.73245279,  0.0542589 ,  0.27129451,  0.89246539,  0.59254869],
            [ 0.39294651,  0.28154534,  0.73245279,  0.0542589 ,  0.27129451,  0.45863938],
            [ 0.44553035,  0.39294651,  0.28154534,  0.73245279,  0.0542589 ,  0.1640389 ]]
        ),
        columns = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 'exog'],
        index = pd.RangeIndex(start=8, stop=13, step=1)
    )
    
    pd.testing.assert_frame_equal(results, expected)


def test_create_predict_X_when_regressor_is_LinearRegression_with_transform_y_and_transform_exog_df():
    """
    Test create_predict_X when using LinearRegression as regressor, StandardScaler
    as transformer_y and transformer_exog as transformer_exog.
    """
    y = pd.Series(
            np.array([-0.59,  0.02, -0.9,  1.09, -3.61,  0.72, -0.11, -0.4])
        )
    exog = pd.DataFrame({
               'col_1': [7.5, 24.4, 60.3, 57.3, 50.7, 41.4, 87.2, 47.4],
               'col_2': ['a', 'a', 'a', 'a', 'b', 'b', 'b', 'b']}
           )
    exog_predict = exog.copy()
    exog_predict.index = pd.RangeIndex(start=8, stop=16)

    transformer_y = StandardScaler()
    transformer_exog = ColumnTransformer(
                            [('scale', StandardScaler(), ['col_1']),
                             ('onehot', OneHotEncoder(), ['col_2'])],
                            remainder = 'passthrough',
                            verbose_feature_names_out = False
                       )
    
    forecaster = ForecasterRecursive(
                     regressor        = LinearRegression(),
                     lags             = 5,
                     transformer_y    = transformer_y,
                     transformer_exog = transformer_exog
                 )
    forecaster.fit(y=y, exog=exog)
    results = forecaster.create_predict_X(steps=5, exog=exog_predict)

    expected = pd.DataFrame(
        data = np.array([[ 0.0542589 ,  0.27129451,  0.89246539, -2.34810076,  1.16937289,
                          -1.76425513,  1.        ,  0.        ],
                         [ 0.73245279,  0.0542589 ,  0.27129451,  0.89246539, -2.34810076,
                          -1.00989936,  1.        ,  0.        ],
                         [ 0.28154534,  0.73245279,  0.0542589 ,  0.27129451,  0.89246539,
                           0.59254869,  1.        ,  0.        ],
                         [ 0.39294651,  0.28154534,  0.73245279,  0.0542589 ,  0.27129451,
                           0.45863938,  1.        ,  0.        ],
                         [ 0.44553035,  0.39294651,  0.28154534,  0.73245279,  0.0542589 ,
                           0.1640389 ,  0.        ,  1.        ]]),
        columns = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 'col_1', 'col_2_a', 'col_2_b'],
        index = pd.RangeIndex(start=8, stop=13, step=1)
    )
    
    pd.testing.assert_frame_equal(results, expected)


def test_create_predict_X_when_categorical_features_native_implementation_HistGradientBoostingRegressor():
    """
    Test create_predict_X when using HistGradientBoostingRegressor and categorical variables.
    """
    df_exog = pd.DataFrame(
        {'exog_1': exog_categorical,
         'exog_2': ['a', 'b', 'c', 'd', 'e'] * 10,
         'exog_3': pd.Categorical(['F', 'G', 'H', 'I', 'J'] * 10)}
    )
    
    exog_predict = df_exog.copy()
    exog_predict.index = pd.RangeIndex(start=50, stop=100)

    categorical_features = df_exog.select_dtypes(exclude=[np.number]).columns.tolist()
    transformer_exog = make_column_transformer(
                           (
                               OrdinalEncoder(
                                   dtype=int,
                                   handle_unknown="use_encoded_value",
                                   unknown_value=-1,
                                   encoded_missing_value=-1
                               ),
                               categorical_features
                           ),
                           remainder="passthrough",
                           verbose_feature_names_out=False,
                       ).set_output(transform="pandas")
    
    forecaster = ForecasterRecursive(
                     regressor        = HistGradientBoostingRegressor(
                                            categorical_features = categorical_features,
                                            random_state         = 123
                                        ),
                     lags             = 5,
                     transformer_y    = None,
                     transformer_exog = transformer_exog
                 )
    forecaster.fit(y=y_categorical, exog=df_exog)
    results = forecaster.create_predict_X(steps=10, exog=exog_predict)

    expected = pd.DataFrame(
        data = np.array([
                [0.61289453, 0.51948512, 0.98555979, 0.48303426, 0.25045537,
                    0.        , 0.        , 0.12062867],
                [0.61187012, 0.61289453, 0.51948512, 0.98555979, 0.48303426,
                    1.        , 1.        , 0.8263408 ],
                [0.42274801, 0.61187012, 0.61289453, 0.51948512, 0.98555979,
                    2.        , 2.        , 0.60306013],
                [0.43214802, 0.42274801, 0.61187012, 0.61289453, 0.51948512,
                    3.        , 3.        , 0.54506801],
                [0.4923281 , 0.43214802, 0.42274801, 0.61187012, 0.61289453,
                    4.        , 4.        , 0.34276383],
                [0.53073262, 0.4923281 , 0.43214802, 0.42274801, 0.61187012,
                    0.        , 0.        , 0.30412079],
                [0.48004443, 0.53073262, 0.4923281 , 0.43214802, 0.42274801,
                    1.        , 1.        , 0.41702221],
                [0.56731689, 0.48004443, 0.53073262, 0.4923281 , 0.43214802,
                    2.        , 2.        , 0.68130077],
                [0.53956024, 0.56731689, 0.48004443, 0.53073262, 0.4923281 ,
                    3.        , 3.        , 0.87545684],
                [0.47670124, 0.53956024, 0.56731689, 0.48004443, 0.53073262,
                    4.        , 4.        , 0.51042234]]),
        columns = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 'exog_2', 'exog_3', 'exog_1'],
        index = pd.RangeIndex(start=50, stop=60, step=1)
    ).astype({'exog_2': int, 'exog_3': int})
    
    pd.testing.assert_frame_equal(results, expected)


def test_create_predict_X_when_categorical_features_auto_detect_LGBMRegressor():
    """
    Test create_predict_X when using LGBMRegressor and categorical variables.
    """
    df_exog = pd.DataFrame(
        {'exog_1': exog_categorical,
         'exog_2': ['a', 'b', 'c', 'd', 'e'] * 10,
         'exog_3': pd.Categorical(['F', 'G', 'H', 'I', 'J'] * 10)}
    )
    
    exog_predict = df_exog.copy()
    exog_predict.index = pd.RangeIndex(start=50, stop=100)

    pipeline_categorical = make_pipeline(
                               OrdinalEncoder(
                                   dtype=int,
                                   handle_unknown="use_encoded_value",
                                   unknown_value=-1,
                                   encoded_missing_value=-1
                               ),
                               FunctionTransformer(
                                   func=lambda x: x.astype('category'),
                                   feature_names_out= 'one-to-one'
                               )
                           )

    transformer_exog = make_column_transformer(
                           (
                               pipeline_categorical,
                               make_column_selector(dtype_exclude=np.number)
                           ),
                           remainder="passthrough",
                           verbose_feature_names_out=False,
                       ).set_output(transform="pandas")
    
    forecaster = ForecasterRecursive(
                     regressor        = LGBMRegressor(verbose=-1, random_state=123),
                     lags             = 5,
                     transformer_y    = None,
                     transformer_exog = transformer_exog
                 )
    forecaster.fit(y=y_categorical, exog=df_exog)
    results = forecaster.create_predict_X(steps=10, exog=exog_predict)

    expected = pd.DataFrame(
        data = np.array([
                [0.61289453, 0.51948512, 0.98555979, 0.48303426, 0.25045537,
                 0.        , 0.        , 0.12062867],
                [0.5857033 , 0.61289453, 0.51948512, 0.98555979, 0.48303426,
                 1.        , 1.        , 0.8263408 ],
                [0.3894503 , 0.5857033 , 0.61289453, 0.51948512, 0.98555979,
                 2.        , 2.        , 0.60306013],
                [0.45053399, 0.3894503 , 0.5857033 , 0.61289453, 0.51948512,
                 3.        , 3.        , 0.54506801],
                [0.49686551, 0.45053399, 0.3894503 , 0.5857033 , 0.61289453,
                 4.        , 4.        , 0.34276383],
                [0.45887492, 0.49686551, 0.45053399, 0.3894503 , 0.5857033 ,
                 0.        , 0.        , 0.30412079],
                [0.51481068, 0.45887492, 0.49686551, 0.45053399, 0.3894503 ,
                 1.        , 1.        , 0.41702221],
                [0.5857033 , 0.51481068, 0.45887492, 0.49686551, 0.45053399,
                 2.        , 2.        , 0.68130077],
                [0.3894503 , 0.5857033 , 0.51481068, 0.45887492, 0.49686551,
                 3.        , 3.        , 0.87545684],
                [0.45053399, 0.3894503 , 0.5857033 , 0.51481068, 0.45887492,
                 4.        , 4.        , 0.51042234]]),
        columns = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 'exog_2', 'exog_3', 'exog_1'],
        index = pd.RangeIndex(start=50, stop=60, step=1)
    ).astype({'exog_2': int, 'exog_3': int}
    ).astype({'exog_2': 'category', 'exog_3': 'category'})
    
    pd.testing.assert_frame_equal(results, expected)


def test_create_predict_X_when_regressor_is_LinearRegression_with_exog_differentiation_is_1_and_transformer_y():
    """
    Test create_predict_X when using LinearRegression as regressor and differentiation=1,
    and transformer_y is StandardScaler.
    """

    end_train = '2003-03-01 23:59:00'

    # Simulated exogenous variable
    rng = np.random.default_rng(9876)
    exog = pd.Series(
        rng.normal(loc=0, scale=1, size=len(data)), index=data.index, name='exog'
    )
    steps = 4

    forecaster = ForecasterRecursive(
                     regressor       = LinearRegression(),
                     lags            = [1, 5],
                     differentiation = 1
                )
    forecaster.fit(y=data.loc[:end_train], exog=exog.loc[:end_train])
    results = forecaster.create_predict_X(steps=steps, exog=exog.loc[end_train:])

    expected = pd.DataFrame(
        data = np.array([
            [ 0.07503713, -0.01018012,  1.16172882],
            [ 0.00345482,  0.10597971,  0.29468848],
            [-0.00296884, -0.01463081, -0.4399757 ],
            [ 0.01758812, -0.48984868,  1.25008389]]
        ),
        columns = ['lag_1', 'lag_5', 'exog'],
        index = pd.date_range(start='2003-04-01', periods=steps, freq='MS')
    )
    
    pd.testing.assert_frame_equal(results, expected)


def test_create_predict_X_when_window_features():
    """
    Test the output of create_predict_X when using window_features and exog 
    with datetime index.
    """
    y_datetime = pd.Series(
        np.arange(15), index=pd.date_range('2000-01-01', periods=15, freq='D'),
        name='y', dtype=float
    )
    exog_datetime = pd.Series(
        np.arange(100, 115), index=pd.date_range('2000-01-01', periods=15, freq='D'),
        name='exog', dtype=float
    )
    exog_datetime_pred = pd.Series(
        np.arange(115, 120), index=pd.date_range('2000-01-16', periods=5, freq='D'),
        name='exog', dtype=float
    )
    rolling = RollingFeatures(stats=['mean', 'median'], window_sizes=[5, 5])
    rolling_2 = RollingFeatures(stats='sum', window_sizes=[6])

    forecaster = ForecasterRecursive(
        LinearRegression(), lags=5, window_features=[rolling, rolling_2]
    )
    forecaster.fit(y=y_datetime, exog=exog_datetime)
    results = forecaster.create_predict_X(steps=5, exog=exog_datetime_pred)

    expected = pd.DataFrame(
        data = np.array([
                    [14., 13., 12., 11., 10., 12., 12., 69., 115.],
                    [15., 14., 13., 12., 11., 13., 13., 75., 116.],
                    [16., 15., 14., 13., 12., 14., 14., 81., 117.],
                    [17., 16., 15., 14., 13., 15., 15., 87., 118.],
                    [18., 17., 16., 15., 14., 16., 16., 93., 119.]]),
        index   = pd.date_range('2000-01-16', periods=5, freq='D'),
        columns = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 
                   'roll_mean_5', 'roll_median_5', 'roll_sum_6', 'exog']
    )

    pd.testing.assert_frame_equal(results, expected)


def test_create_predict_X_when_window_features_and_lags_None():
    """
    Test the output of create_predict_X when using window_features and exog 
    with datetime index and lags=None.
    """
    y_datetime = pd.Series(
        np.arange(15), index=pd.date_range('2000-01-01', periods=15, freq='D'),
        name='y', dtype=float
    )
    exog_datetime = pd.Series(
        np.arange(100, 115), index=pd.date_range('2000-01-01', periods=15, freq='D'),
        name='exog', dtype=float
    )
    exog_datetime_pred = pd.Series(
        np.arange(115, 120), index=pd.date_range('2000-01-16', periods=5, freq='D'),
        name='exog', dtype=float
    )
    rolling = RollingFeatures(stats=['mean', 'median'], window_sizes=[5, 5])
    rolling_2 = RollingFeatures(stats='sum', window_sizes=[6])

    forecaster = ForecasterRecursive(
        LinearRegression(), lags=None, window_features=[rolling, rolling_2]
    )
    forecaster.fit(y=y_datetime, exog=exog_datetime)
    results = forecaster.create_predict_X(steps=5, exog=exog_datetime_pred)

    expected = pd.DataFrame(
        data = np.array([
                    [12., 12., 69., 115.],
                    [13., 13., 75., 116.],
                    [14., 14., 81., 117.],
                    [15., 15., 87., 118.],
                    [16., 16., 93., 119.]]),
        index   = pd.date_range('2000-01-16', periods=5, freq='D'),
        columns = ['roll_mean_5', 'roll_median_5', 'roll_sum_6', 'exog']
    )

    pd.testing.assert_frame_equal(results, expected)


def test_create_predict_X_same_predictions_as_predict():
    """
    Test create_predict_X matrix returns the same predictions as predict method
    when passing to the regressor predict method.
    """

    end_train = '2003-03-01 23:59:00'

    # Simulated exogenous variable
    rng = np.random.default_rng(9876)
    exog = pd.Series(
        rng.normal(loc=0, scale=1, size=len(data)), index=data.index, name='exog'
    )
    rolling = RollingFeatures(stats=['mean', 'median'], window_sizes=[5, 5])
    rolling_2 = RollingFeatures(stats='sum', window_sizes=[6])
    steps = 4

    forecaster = ForecasterRecursive(
                     regressor        = LinearRegression(),
                     lags             = [1, 5],
                     window_features  = [rolling, rolling_2],
                     transformer_y    = None,
                     transformer_exog = None,
                     differentiation  = None
                 )
    forecaster.fit(y=data.loc[:end_train], exog=exog.loc[:end_train])
    X_predict = forecaster.create_predict_X(steps=steps, exog=exog.loc[end_train:])
    results = forecaster.regressor.predict(X_predict)

    expected = forecaster.predict(steps=steps, exog=exog.loc[end_train:]).to_numpy()
    
    np.testing.assert_array_almost_equal(results, expected, decimal=7)


def test_create_predict_X_same_predictions_as_predict_transformers():
    """
    Test create_predict_X matrix returns the same predictions as predict method
    when passing to the regressor predict method with transformation.
    """

    end_train = '2003-03-01 23:59:00'

    # Simulated exogenous variable
    rng = np.random.default_rng(9876)
    exog = pd.Series(
        rng.normal(loc=0, scale=1, size=len(data)), index=data.index, name='exog'
    )
    rolling = RollingFeatures(stats=['mean', 'median'], window_sizes=[5, 5])
    rolling_2 = RollingFeatures(stats='sum', window_sizes=[6])
    steps = 4

    forecaster = ForecasterRecursive(
                     regressor        = LinearRegression(),
                     lags             = [1, 5],
                     window_features  = [rolling, rolling_2],
                     transformer_y    = StandardScaler(),
                     transformer_exog = StandardScaler(),
                     differentiation  = None
                 )
    forecaster.fit(y=data.loc[:end_train], exog=exog.loc[:end_train])

    warn_msg = re.escape(
        "The output matrix is in the transformed scale due to the "
        "inclusion of transformations or differentiation in the Forecaster. "
        "As a result, any predictions generated using this matrix will also "
        "be in the transformed scale. Please refer to the documentation "
        "for more details: "
        "https://skforecast.org/latest/user_guides/training-and-prediction-matrices.html"
    )
    with pytest.warns(DataTransformationWarning, match = warn_msg):
        X_predict = forecaster.create_predict_X(steps=steps, exog=exog.loc[end_train:])
    results = forecaster.regressor.predict(X_predict)

    results = transform_numpy(
                  array             = results,
                  transformer       = forecaster.transformer_y,
                  fit               = False,
                  inverse_transform = True
              )

    expected = forecaster.predict(steps=steps, exog=exog.loc[end_train:]).to_numpy()
    
    np.testing.assert_array_almost_equal(results, expected, decimal=7)


def test_create_predict_X_same_predictions_as_predict_transformers_diff():
    """
    Test create_predict_X matrix returns the same predictions as predict method
    when passing to the regressor predict method with transformation and differentiation.
    """

    end_train = '2003-03-01 23:59:00'

    # Simulated exogenous variable
    rng = np.random.default_rng(9876)
    exog = pd.Series(
        rng.normal(loc=0, scale=1, size=len(data)), index=data.index, name='exog'
    )
    rolling = RollingFeatures(stats=['mean', 'median'], window_sizes=[5, 5])
    rolling_2 = RollingFeatures(stats='sum', window_sizes=[6])
    steps = 4

    forecaster = ForecasterRecursive(
                     regressor        = LinearRegression(),
                     lags             = [1, 5],
                     window_features  = [rolling, rolling_2],
                     transformer_y    = StandardScaler(),
                     transformer_exog = StandardScaler(),
                     differentiation  = 1
                 )
    forecaster.fit(y=data.loc[:end_train], exog=exog.loc[:end_train])

    warn_msg = re.escape(
        "The output matrix is in the transformed scale due to the "
        "inclusion of transformations or differentiation in the Forecaster. "
        "As a result, any predictions generated using this matrix will also "
        "be in the transformed scale. Please refer to the documentation "
        "for more details: "
        "https://skforecast.org/latest/user_guides/training-and-prediction-matrices.html"
    )
    with pytest.warns(DataTransformationWarning, match = warn_msg):
        X_predict = forecaster.create_predict_X(steps=steps, exog=exog.loc[end_train:])
    results = forecaster.regressor.predict(X_predict)

    results = forecaster.differentiator.inverse_transform_next_window(results)
    results = transform_numpy(
                  array             = results,
                  transformer       = forecaster.transformer_y,
                  fit               = False,
                  inverse_transform = True
              )

    expected = forecaster.predict(steps=steps, exog=exog.loc[end_train:]).to_numpy()
    
    np.testing.assert_array_almost_equal(results, expected, decimal=7)



# Unit test backtesting_forecaster
# ==============================================================================
import re
import pytest
from sklearn.linear_model import Ridge
from skforecast.recursive import ForecasterRecursiveMultiSeries
from skforecast.model_selection import backtesting_forecaster
from skforecast.model_selection._split import TimeSeriesFold

# Fixtures
from ..fixtures_model_selection import y


def test_backtesting_forecaster_TypeError_when_forecaster_not_supported_types():
    """
    Test TypeError is raised in backtesting_forecaster if Forecaster is not one 
    of the types supported by the function.
    """
    forecaster = ForecasterRecursiveMultiSeries(
                    regressor = Ridge(random_state=123),
                    lags      = 2
                 )
    forecaters_allowed = [
        'ForecasterRecursive', 
        'ForecasterDirect',
        'ForecasterEquivalentDate'
    ]
    cv = TimeSeriesFold(
            steps                 = 3,
            initial_train_size    = len(y[:-12]),
            window_size           = None,
            differentiation       = None,
            refit                 = False,
            fixed_train_size      = True,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    err_msg = re.escape(
        f"`forecaster` must be of type {forecaters_allowed}, for all other types of "
        f" forecasters use the functions available in the other `model_selection` "
        f"modules."
    )
    
    with pytest.raises(TypeError, match = err_msg):
        backtesting_forecaster(
            forecaster            = forecaster,
            y                     = y,
            cv                    = cv,
            metric                = 'mean_absolute_error',
            exog                  = None,
            verbose               = False,
            show_progress         = False
        )



# Unit test predict_interval ForecasterSarimax
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from sklearn.exceptions import NotFittedError
from skforecast.sarimax import Sarimax
from skforecast.recursive import ForecasterSarimax
from skforecast.utils import expand_index
from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import StandardScaler
from sklearn.preprocessing import OneHotEncoder

# Fixtures
from .fixtures_forecaster_sarimax import y
from .fixtures_forecaster_sarimax import y_lw
from .fixtures_forecaster_sarimax import exog
from .fixtures_forecaster_sarimax import exog_lw
from .fixtures_forecaster_sarimax import exog_predict
from .fixtures_forecaster_sarimax import exog_lw_predict
from .fixtures_forecaster_sarimax import y_datetime
from .fixtures_forecaster_sarimax import y_lw_datetime
from .fixtures_forecaster_sarimax import exog_datetime
from .fixtures_forecaster_sarimax import exog_lw_datetime
from .fixtures_forecaster_sarimax import exog_predict_datetime
from .fixtures_forecaster_sarimax import exog_lw_predict_datetime
from .fixtures_forecaster_sarimax import df_exog
from .fixtures_forecaster_sarimax import df_exog_lw
from .fixtures_forecaster_sarimax import df_exog_predict
from .fixtures_forecaster_sarimax import df_exog_lw_predict
from .fixtures_forecaster_sarimax import df_exog_datetime
from .fixtures_forecaster_sarimax import df_exog_lw_datetime
from .fixtures_forecaster_sarimax import df_exog_predict_datetime
from .fixtures_forecaster_sarimax import df_exog_lw_predict_datetime


def test_predict_NotFittedError_when_fitted_is_False():
    """
    Test NotFittedError is raised when fitted is False.
    """
    forecaster = ForecasterSarimax(regressor=Sarimax(order=(1, 1, 1)))

    err_msg = re.escape(
        ("This Forecaster instance is not fitted yet. Call `fit` with "
         "appropriate arguments before using predict.")
    )
    with pytest.raises(NotFittedError, match = err_msg):
        forecaster.predict_interval(
            steps = 5, 
            alpha = 0.05
        )


def test_predict_interval_ValueError_when_ForecasterSarimax_last_window_exog_is_not_None_and_last_window_is_not_provided():
    """
    Check ValueError is raised when last_window_exog is not None, but 
    last_window is not provided.
    """
    forecaster = ForecasterSarimax(regressor=Sarimax(order=(1, 1, 1)))
    forecaster.fit(y=y, exog=exog)
    
    err_msg = re.escape(
                ("To make predictions unrelated to the original data, both "
                 "`last_window` and `last_window_exog` must be provided.")
              )   
    with pytest.raises(ValueError, match = err_msg):
        forecaster.predict_interval(
            steps            = 5, 
            alpha            = 0.05, 
            exog             = exog_predict, 
            last_window      = None, 
            last_window_exog = exog
        )


def test_predict_interval_ValueError_when_ForecasterSarimax_last_window_exog_is_None_and_included_exog_is_true():
    """
    Check ValueError is raised when last_window_exog is None, but included_exog
    is True and last_window is provided.
    """
    forecaster = ForecasterSarimax(regressor=Sarimax(order=(1, 1, 1)))
    forecaster.fit(y=y, exog=exog)
    
    err_msg = re.escape(
                ("Forecaster trained with exogenous variable/s. To make predictions "
                 "unrelated to the original data, same variable/s must be provided "
                 "using `last_window_exog`.")
              )   
    with pytest.raises(ValueError, match = err_msg):
        forecaster.predict_interval(
            steps            = 5, 
            alpha            = 0.05, 
            exog             = exog_lw_predict, 
            last_window      = y_lw, 
            last_window_exog = None
        )


def test_predict_interval_ValueError_when_interval_is_not_symmetrical():
    """
    Raise ValueError if `interval` is not symmetrical.
    """
    forecaster = ForecasterSarimax(regressor=Sarimax(order=(1, 1, 1)))
    forecaster.fit(y=y)
    alpha = None
    interval_not_symmetrical = [5, 97.5] 

    err_msg = re.escape(
                (f"When using `interval` in ForecasterSarimax, it must be symmetrical. "
                 f"For example, interval of 95% should be as `interval = [2.5, 97.5]`. "
                 f"Got {interval_not_symmetrical}.")
            )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.predict_interval(
            steps    = 5, 
            alpha    = alpha, 
            interval = interval_not_symmetrical
        )


@pytest.mark.parametrize("alpha, interval", 
                         [(0.05, [1, 99]), 
                          (None, [2.5, 97.5])], 
                         ids = lambda values : f'alpha, interval: {values}')
def test_predict_interval_output_ForecasterSarimax_skforecast_Sarimax(alpha, interval):
    """
    Test predict_interval output of ForecasterSarimax using Sarimax from skforecast.
    """
    forecaster = ForecasterSarimax(
                     regressor = Sarimax(maxiter=2000, method='cg', disp=False, order=(3, 2, 0))
                 )
    forecaster.fit(y=y)
    predictions = forecaster.predict_interval(steps=5, alpha=alpha, interval=interval)
    expected = pd.DataFrame(
                   data    = np.array([[ 0.53809981,  0.24275351,  0.83344611],
                                       [ 0.53145374,  0.0431938 ,  1.01971368],
                                       [ 0.53763636, -0.12687285,  1.20214556],
                                       [ 0.52281442, -0.35748984,  1.40311868],
                                       [ 0.49770378, -0.64436866,  1.63977622]]),
                   columns = ['pred', 'lower_bound', 'upper_bound'],
                   index   = pd.RangeIndex(start=50, stop=55, step=1)
               )
    
    pd.testing.assert_frame_equal(predictions, expected)


@pytest.mark.parametrize("alpha, interval", 
                         [(0.05, [1, 99]), 
                          (None, [2.5, 97.5])], 
                         ids = lambda values : f'alpha, interval: {values}')
def test_predict_interval_output_ForecasterSarimax_with_exog(alpha, interval):
    """
    Test predict_interval output of ForecasterSarimax with exogenous variables.
    """
    forecaster = ForecasterSarimax(
                     regressor = Sarimax(maxiter=1000, method='cg', disp=False, order=(1, 0, 1))
                 )
    forecaster.fit(y=y, exog=exog)
    predictions = forecaster.predict_interval(steps=5, exog=exog_predict, alpha=alpha, interval=interval)
    expected = pd.DataFrame(
                   data    = np.array([[0.59929905, 0.57862017, 0.61997793],
                                       [0.61299725, 0.59202539, 0.63396911],
                                       [0.6287311 , 0.60774224, 0.64971995],
                                       [0.64413557, 0.62314573, 0.66512542],
                                       [0.66195978, 0.64096988, 0.68294969]]),
                   columns = ['pred', 'lower_bound', 'upper_bound'],
                   index   = pd.RangeIndex(start=50, stop=55, step=1)
               )
    
    pd.testing.assert_frame_equal(predictions, expected, atol=0.0001)


@pytest.mark.parametrize("alpha, interval", 
                         [(0.05, [1, 99]), 
                          (None, [2.5, 97.5])], 
                         ids = lambda values : f'alpha, interval: {values}')
def test_predict_interval_output_ForecasterSarimax_with_transform_y(alpha, interval):
    """
    Test predict_interval output of ForecasterSarimax with a StandardScaler() as transformer_y.
    """
    forecaster = ForecasterSarimax(
                     regressor     = Sarimax(maxiter=1000, method='cg', disp=False, order=(1, 1, 1)),
                     transformer_y = StandardScaler()
                 )
    forecaster.fit(y=y)
    predictions = forecaster.predict_interval(steps=5, alpha=alpha, interval=interval)
    expected = pd.DataFrame(
                   data    = np.array([[0.63520867, 0.61383185, 0.6565855 ],
                                       [0.61741115, 0.5894499 , 0.6453724 ],
                                       [0.6330291 , 0.60053638, 0.66552182],
                                       [0.6193238 , 0.58402618, 0.65462142],
                                       [0.63135068, 0.59379186, 0.66890951]]),
                   columns = ['pred', 'lower_bound', 'upper_bound'],
                   index   = pd.RangeIndex(start=50, stop=55, step=1)
               )
    
    pd.testing.assert_frame_equal(predictions, expected, atol=0.0001)


@pytest.mark.parametrize("alpha, interval", 
                         [(0.05, [1, 99]), 
                          (None, [2.5, 97.5])], 
                         ids = lambda values : f'alpha, interval: {values}')
def test_predict_interval_output_ForecasterSarimax_with_transform_y_and_transform_exog(alpha, interval):
    """
    Test predict_interval output of ForecasterSarimax, StandardScaler
    as transformer_y and transformer_exog as transformer_exog.
    """
    transformer_exog = ColumnTransformer(
                           [('scale', StandardScaler(), ['exog_1']),
                            ('onehot', OneHotEncoder(), ['exog_2'])],
                           remainder = 'passthrough',
                           verbose_feature_names_out = False
                       )

    forecaster = ForecasterSarimax(
                     regressor        = Sarimax(maxiter=1000, method='cg', disp=False, order=(1, 0, 1)),
                     transformer_y    = StandardScaler(),
                     transformer_exog = transformer_exog
                 )
    forecaster.fit(y=y, exog=df_exog)
    predictions = forecaster.predict_interval(steps=5, exog=df_exog_predict, alpha=alpha, interval=interval)
    expected = pd.DataFrame(
                   data    = np.array([[0.60687311, 0.50667956, 0.70706666],
                                       [0.62484493, 0.49759696, 0.75209289],
                                       [0.63515416, 0.50776733, 0.762541  ],
                                       [0.67730912, 0.54992148, 0.80469675],
                                       [0.69458838, 0.56720074, 0.82197602]]),
                   columns = ['pred', 'lower_bound', 'upper_bound'],
                   index   = pd.RangeIndex(start=50, stop=55, step=1)
               )
    
    pd.testing.assert_frame_equal(predictions, expected, atol=0.0001)


def test_predict_interval_ValueError_when_last_window_index_does_not_follow_training_set():
    """
    Raise ValueError if `last_window` index does not start at the end 
    of the index seen by the forecaster.
    """
    y_test = pd.Series(data=y_datetime.to_numpy())
    y_test.index = pd.date_range(start='2022-01-01', periods=50, freq='D')
    lw_test = pd.Series(data=y_lw_datetime.to_numpy())
    lw_test.index = pd.date_range(start='2022-03-01', periods=50, freq='D')

    forecaster = ForecasterSarimax(regressor = Sarimax(order=(1, 0, 0)))
    forecaster.fit(y=y_test)
    expected_index = expand_index(forecaster.extended_index_, 1)[0]

    err_msg = re.escape(
        (f"To make predictions unrelated to the original data, `last_window` "
         f"has to start at the end of the index seen by the forecaster.\n"
         f"    Series last index         : {forecaster.extended_index_[-1]}.\n"
         f"    Expected index            : {expected_index}.\n"
         f"    `last_window` index start : {lw_test.index[0]}.")
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.predict_interval(
            steps            = 5, 
            alpha            = 0.05,
            last_window      = lw_test,
        )


def test_predict_interval_ValueError_when_last_window_exog_index_does_not_follow_training_set():
    """
    Raise ValueError if `last_window_exog` index does not start at the end 
    of the index seen by the forecaster.
    """
    y_test = pd.Series(data=y_datetime.to_numpy())
    y_test.index = pd.date_range(start='2022-01-01', periods=50, freq='D')
    lw_test = pd.Series(data=y_lw_datetime.to_numpy())
    lw_test.index = pd.date_range(start='2022-02-20', periods=50, freq='D')

    exog_test = pd.Series(data=exog_datetime.to_numpy(), name='exog')
    exog_test.index = pd.date_range(start='2022-01-01', periods=50, freq='D')
    exog_pred_test = pd.Series(data=exog_predict_datetime.to_numpy(), name='exog')
    exog_pred_test.index = pd.date_range(start='2022-04-11', periods=10, freq='D')
    lw_exog_test = pd.Series(data=exog_lw_datetime.to_numpy(), name='exog')
    lw_exog_test.index = pd.date_range(start='2022-03-01', periods=50, freq='D')

    forecaster = ForecasterSarimax(regressor = Sarimax(order=(1, 0, 0)))
    forecaster.fit(y=y_test, exog=exog_test)
    expected_index = expand_index(forecaster.extended_index_, 1)[0]

    err_msg = re.escape(
        (f"To make predictions unrelated to the original data, `last_window_exog` "
         f"has to start at the end of the index seen by the forecaster.\n"
         f"    Series last index              : {forecaster.extended_index_[-1]}.\n"
         f"    Expected index                 : {expected_index}.\n"
         f"    `last_window_exog` index start : {lw_exog_test.index[0]}.")
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.predict_interval(
            steps            = 5, 
            alpha            = 0.05,
            exog             = exog_pred_test, 
            last_window      = lw_test, 
            last_window_exog = lw_exog_test
        )


@pytest.mark.parametrize("alpha, interval", 
                         [(0.05, [1, 99]), 
                          (None, [2.5, 97.5])], 
                         ids = lambda values : f'alpha, interval: {values}')
def test_predict_interval_output_ForecasterSarimax_with_last_window(alpha, interval):
    """
    Test predict_interval output of ForecasterSarimax with `last_window`.
    """
    forecaster = ForecasterSarimax(
                     regressor = Sarimax(maxiter=1000, method='cg', disp=False, order=(3, 2, 0))
                 )
    forecaster.fit(y=y_datetime)
    predictions = forecaster.predict_interval(
                      steps            = 5, 
                      alpha            = alpha,
                      interval         = interval,
                      last_window      = y_lw_datetime,
                  )
    
    expected = pd.DataFrame(
                    data = np.array([[0.91877817, 0.62343187, 1.21412446],
                                     [0.98433512, 0.49607518, 1.47259506],
                                     [1.06945921, 0.40495001, 1.73396842],
                                     [1.15605055, 0.27574629, 2.03635481],
                                     [1.22975713, 0.08768469, 2.37182957]]),
                   columns = ['pred', 'lower_bound', 'upper_bound'],
                   index   = pd.date_range(start='2100', periods=5, freq='YE')
               )
    
    pd.testing.assert_frame_equal(predictions, expected, atol=0.0001)


@pytest.mark.parametrize("alpha, interval", 
                         [(0.05, [1, 99]), 
                          (None, [2.5, 97.5])], 
                         ids = lambda values : f'alpha, interval: {values}')
def test_predict_interval_output_ForecasterSarimax_with_last_window_and_exog(alpha, interval):
    """
    Test predict_interval output of ForecasterSarimax with exogenous variables and `last_window`.
    """
    forecaster = ForecasterSarimax(
                     regressor = Sarimax(maxiter=1000, method='cg', disp=False, order=(1, 1, 1))
                 )
    forecaster.fit(y=y_datetime, exog=exog_datetime)
    predictions = forecaster.predict_interval(
                      steps            = 5, 
                      alpha            = alpha,
                      interval         = interval,
                      exog             = exog_lw_predict_datetime, 
                      last_window      = y_lw_datetime, 
                      last_window_exog = exog_lw_datetime
                  )

    expected = pd.DataFrame(
                   data    = np.array([[ 0.89386888, -0.84405923,  2.63179699],
                                       [ 0.92919515, -1.45638221,  3.3147725 ],
                                       [ 0.98327241, -1.88128514,  3.84782996],
                                       [ 1.02336286, -2.2399583 ,  4.28668401],
                                       [ 1.05334974, -2.56051157,  4.66721105]]),
                   columns = ['pred', 'lower_bound', 'upper_bound'],
                   index   = pd.date_range(start='2100', periods=5, freq='YE')
               )
    
    pd.testing.assert_frame_equal(predictions, expected, atol=0.0001)


@pytest.mark.parametrize("alpha, interval", 
                         [(0.05, [1, 99]), 
                          (None, [2.5, 97.5])], 
                         ids = lambda values : f'alpha, interval: {values}')
def test_predict_interval_output_ForecasterSarimax_with_last_window_and_exog_and_transformers(alpha, interval):
    """
    Test predict_interval output of ForecasterSarimax with exogenous variables and `last_window`.
    """
    transformer_exog = ColumnTransformer(
                           [('scale', StandardScaler(), ['exog_1']),
                            ('onehot', OneHotEncoder(), ['exog_2'])],
                           remainder = 'passthrough',
                           verbose_feature_names_out = False
                       )

    forecaster = ForecasterSarimax(
                     regressor = Sarimax(maxiter=1000, method='cg', disp=False, order=(1, 1, 1)),
                     transformer_y    = StandardScaler(),
                     transformer_exog = transformer_exog
                 )
    forecaster.fit(y=y_datetime, exog=df_exog_datetime)
    predictions = forecaster.predict_interval(
                      steps            = 5, 
                      alpha            = alpha,
                      interval         = interval,
                      exog             = df_exog_lw_predict_datetime, 
                      last_window      = y_lw_datetime, 
                      last_window_exog = df_exog_lw_datetime
                  )

    expected = pd.DataFrame(
                   data    = np.array([[0.61139264, 0.35457567, 0.86820961],
                                       [0.88228163, 0.57163268, 1.19293057],
                                       [0.77749663, 0.42990006, 1.12509319],
                                       [0.94985823, 0.58885008, 1.31086638],
                                       [0.89218798, 0.5184476 , 1.26592836]]),
                   columns = ['pred', 'lower_bound', 'upper_bound'],
                   index   = pd.date_range(start='2100', periods=5, freq='YE')
               )
    
    pd.testing.assert_frame_equal(predictions, expected, atol=0.0001)


@pytest.mark.parametrize("y          , idx", 
                         [(y         , pd.RangeIndex(start=0, stop=50)), 
                          (y_datetime, pd.date_range(start='2000', periods=50, freq='YE'))], 
                         ids = lambda values: f'y, index: {type(values)}')
def test_predict_interval_ForecasterSarimax_updates_extended_index_twice(y, idx):
    """
    Test forecaster.extended_index_ is updated when using predict_interval twice.
    """
    y_fit = y.iloc[:30].copy()

    forecaster = ForecasterSarimax(regressor=Sarimax(order=(1, 0, 0)))
    forecaster.fit(y=y_fit)

    lw_1 = y.iloc[30:40].copy()
    forecaster.predict_interval(steps=5, alpha = 0.05, last_window=lw_1)
    result_1 = forecaster.extended_index_.copy()
    expected_1 = idx[:40]

    lw_2 = y.iloc[40:].copy()
    forecaster.predict_interval(steps=5, alpha = 0.05, last_window=lw_2)

    pd.testing.assert_index_equal(result_1, expected_1)
    pd.testing.assert_index_equal(forecaster.extended_index_, idx)
